////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: ceresexport.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "util/xmlwrite.h"
#include "util/proofdbdtd.h"
#include "term_operations.h"
#include "libproof.h"

#include "typehlk.h"
#include "compilehlk.h"
#include "symbolhlk.h"
#include "ceresexport.h"

#include <iostream>//temp
using namespace std;
using namespace ProofLib;

////////////////////////////////////////////////////////////////////////////////

#if 0
class ExportCeresImpl;

ExportCeresImpl& operator << ( ExportCeresImpl& o, const ProofLink& p );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Variable& v );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Constant& c );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Function& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Term& t );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const StaticTermList& l );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const AtomFormula& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const QuantorFormula& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const UnaryConFormula& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const BinaryConFormula& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const FormulaVariable& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Formula& f );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const FormulaListVariable& l );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const FormulaListElement& e );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const SElement& e );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const SElementList& l );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Sequent& s );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Rule& r );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const RuleItem& ruleitem );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Proof& p );
ExportCeresImpl& operator << ( ExportCeresImpl& o, const ProofDatabase& pdb );
#endif

////////////////////////////////////////////////////////////////////////////////

class ExportCeresImpl : public ExportCeres
{
  XmlWriter             *m_xml_writer;
  int                    m_compression;

  const ProofDatabase   *m_pdb;

protected:
  ExportCeresImpl()
  {
    m_xml_writer = NULL;
    m_compression = 0;
    m_pdb = NULL;
  }

  friend class ExportCeres;

public:
  XmlWriter* GetXmlWriter() const
  {
    return m_xml_writer;
  }

  const ProofDatabase* GetProofDatabase() const
  {
    return m_pdb;
  }

  virtual void Write( const ProofLib::ProofDatabase* proofdb,
                      const HLK::UsedDefinitionsList *defs,
                      const HLK::Document* doc )
  {
    ASSERT( doc != NULL );
    ASSERT( m_xml_writer != NULL );

    m_xml_writer->StartDocument( PROOFDBDTD_VERSION, PROOFDBDTD_ENCODING );

    m_xml_writer->WriteDTD( PROOFDBDTD_NODE_PROOFDATABASE, NULL,
                            PROOFDBDTD_DTD_PROOFDATABASE, NULL );


    m_xml_writer->OpenElement( PROOFDBDTD_NODE_PROOFDATABASE );

    Write_UsedDefinitionsList( defs, doc );

    m_xml_writer->OpenElement( PROOFDBDTD_NODE_AXIOMSET );

    HLK::AxiomTable *at = doc->GetAxiomTable();

    for( size_t i = 0; i < at->GetCount(); i++ )
      ((ExportXML&) *this) << *(at->GetAxiomAt(i));

    m_xml_writer->CloseElement();

    Write_( proofdb );

    m_xml_writer->OpenElement( PROOFDBDTD_NODE_VARIABLE_DEFINITIONS );
    m_xml_writer->CloseElement();

    m_xml_writer->CloseElement();

    m_xml_writer->EndDocument();
  }

  virtual void Write_UsedDefinitionsList( const HLK::UsedDefinitionsList *defs,
                                          const HLK::Document* doc )
  {
    m_xml_writer->OpenElement( PROOFDBDTD_NODE_DEFINITIONLIST );

    for( size_t i = 0; i < defs->GetCount(); i++ )
    {
      const ProofLib::Symbol *sym = defs->GetSymbol(i);

      std::set<size_t> levels = defs->GetRecursionLevels(i);
      std::set<size_t>::const_iterator j = levels.begin();

      for( ; j != levels.end(); j++ )
      {
        Write_UsedDefinition( sym, *j, doc );
      }
    }

    HLK::SymbolTable *st = doc->GetSymbolTable();

    for( size_t i = 0; i < st->GetEqualityDefinitionCount(); i++ )
    {
      const HLK::TypeSymbol *l, *r;
      USTRING id;
      ProofLib::Formula* f = st->GetEqualityDefinitionAt( i, &l, &r, &id );

      m_xml_writer->OpenElement( PROOFDBDTD_NODE_FORMULADEF );

/*
      std::wstringstream str;
      str << L"=";
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL, str.str().c_str() );
      str << L"_(" << l->GetName() << L"x" << r->GetName() << L")";
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_ID, str.str().c_str() );
*/
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL, L"=" );
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_ID, id );

#if 0
      std::cerr << sym->GetName() << ", " << level;
      std::cerr << std::endl;
#endif

      ProofLib::TermList vars;
      ProofLib::OperationGetFreeVariables op( vars );
      f->Apply( op );
      Write_VariableList( &vars );

      ((ExportXML&) *this) << *f;

      m_xml_writer->CloseElement();
    }

    m_xml_writer->CloseElement();
  }

  virtual void Write_UsedDefinition( const ProofLib::Symbol *sym, size_t level,
                                     const HLK::Document* doc )
  {
    const HLK::PredicateSymbol *ps =
      dynamic_cast<const HLK::PredicateSymbol*>(sym);

    if( ps != NULL )
    {
      if( ps->IsEqualPred() )
        return; // IGNORE EQUALITY DEFINITIONS FOR NOW!

      m_xml_writer->OpenElement( PROOFDBDTD_NODE_FORMULADEF );

      std::wstringstream str;
      str << sym->GetName();
      if( ps->IsRecursive() )
        str << L"[" <<  level << L"]";
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL, str.str().c_str() );
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_ID, str.str().c_str() );

#if 0
      std::cerr << sym->GetName() << ", " << level;
      std::cerr << std::endl;
#endif

      Write_VariableList( ps->GetVariableMapping() );

      if( ps->IsRecursive() )
      {
        ProofLib::Formula* f = ps->GetFormulaAtLevel( level,
          doc->GetNodeDatabase(), doc->GetSymbolTable(), *ps );
        ((ExportXML&) *this) << *f;
        f->Detach();
      }
      else
      {
        ((ExportXML&) *this) << *(ps->GetDefinedFormula());
      }

      m_xml_writer->CloseElement();
      return;
    }

    const HLK::FunctionSymbol *fs =
      dynamic_cast<const HLK::FunctionSymbol*>(sym);

    if( fs != NULL )
    {
      if( fs->GetThroughFormula() != NULL )
        m_xml_writer->OpenElement( PROOFDBDTD_NODE_INDIRECTTERMDEF );
      else
        m_xml_writer->OpenElement( PROOFDBDTD_NODE_TERMDEF );
      std::wstringstream str;
      str << sym->GetName();
      if( fs->IsRecursive() )
        str << L"[" <<  level << L"]";
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL, str.str().c_str() );
      m_xml_writer->WriteAttribute( PROOFDBDTD_ATTR_ID, str.str().c_str() );

#if 0
      std::cerr << "func: " << sym->GetName() << ", " << level;
      std::cerr << std::endl;
#endif

      Write_VariableList( fs->GetVariableMapping() );

      if( fs->GetThroughFormula() != NULL )
      {
        ASSERT( !fs->IsRecursive() );
        ((ExportXML&) *this) << *(fs->GetThroughFormula());
        ((ExportXML&) *this) << *(fs->GetByFormula());
      }
      else
      {
        if( fs->IsRecursive() )
        {
          ProofLib::Term* t = fs->GetTermAtLevel( level,
            doc->GetNodeDatabase(), doc->GetSymbolTable(), *fs );
          ((ExportXML&) *this) << *t;
          t->Detach();
        }
        else
        {
          ((ExportXML&) *this) << *(fs->GetFunctionDefinition());
        }
      }

      m_xml_writer->CloseElement();
      return;
    }
  }
  
  virtual void Write_VariableList( const ProofLib::TermListBase* list )
  {
    m_xml_writer->OpenElement( PROOFDBDTD_NODE_VARIABLELIST );

    if( list != NULL ) for( size_t i = 0; i < list->GetChildCount(); i++ )
    {
      ((ExportXML&) *this) << *(list->GetTerm(i));
    }

    m_xml_writer->CloseElement();
  }

  virtual void Write( const ProofLib::ProofDatabase* proofdb )
  {
    ASSERT( false );
  }

  virtual void Write( const ProofLib::Proof* proof,
                      const ProofLib::ProofDatabase* proofdb )
  {
    ASSERT( false );
  }

  virtual void Write_( const ProofLib::ProofDatabase* proofdb )
  {
    Write_( proofdb->GetDefaultProof(), proofdb );
  }

  virtual void Write_( const ProofLib::Proof* proof,
                       const ProofLib::ProofDatabase* proofdb )
  {
    if( proof == NULL )
      return;

    ASSERT( proofdb != NULL );

    m_pdb = proofdb;

    ((ExportXML&) *this) << *proof;
  }

  virtual void Open( const char *sourcename )
  {
    ASSERT( m_xml_writer == NULL );

    m_xml_writer = XmlWriter::New();

    m_xml_writer->SetCompression( m_compression );
    m_xml_writer->Open( sourcename );
  }

  virtual void Close()
  {
    ASSERT( m_xml_writer != NULL );

    m_xml_writer->Close();
    delete m_xml_writer;
    m_xml_writer = NULL;
  }

  virtual void AddFormulaVar( const ProofLib::FormulaVariable *v ) {}
  virtual void AddFormulaListVar( const ProofLib::FormulaListVariable *v ) {}
  virtual void WriteVariables() {}

  virtual ExportXML& operator << ( const ProofLib::ProofLink& p );
  virtual ExportXML& operator << ( const ProofLib::FormulaVariable& f );
  virtual ExportXML& operator << ( const ProofLib::FormulaListVariable& l );
};

////////////////////////////////////////////////////////////////////////////////

ExportCeres::ExportCeres()
{
}

ExportCeres* ExportCeres::New()
{
  return new ExportCeresImpl();
}

////////////////////////////////////////////////////////////////////////////////

ExportXML& ExportCeresImpl::operator << ( const ProofLink& pl )
{
  Proof *p = pl.GetLinkedProof();
  ASSERT( p != NULL );
  ((ExportXML&) *this) << *(p->GetEndRule());

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportXML& ExportCeresImpl::operator << ( const FormulaVariable& f )
{
  ((ExportXML&) *this) << *(f.GetFormula());

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportXML& ExportCeresImpl::operator << ( const FormulaListVariable& l )
{
#if 0
  if( &l != NULL )
    for( size_t i = 0; i < l.GetChildCount(); i++ )
      ((ExportXML&) *this) << *(l.GetElement(i));
#endif
  ((ExportXML&) *this) << *(l.GetFormulaList());

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

#if 0
ExportCeresImpl& operator << ( ExportCeresImpl& o, const Variable& v )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_VARIABLE );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                    v.GetName() );

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Constant& c )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_CONSTANT );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                    c.GetName() );

  o.GetXmlWriter()->CloseElement();


  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Function& f )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_FUNCTION );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                    f.GetName() );

  o << dynamic_cast<const StaticTermList&>( f );

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Term& t )
{
  try
  {
    const Variable &a = dynamic_cast<const Variable&>( t );
    o << a;
    return o;
  }
  catch( bad_cast ) {}
  
  try
  {
    const Constant &a = dynamic_cast<const Constant&>( t );
    o << a;
    return o;
  }
  catch( bad_cast ) {}
  
  try
  {
    const Function &a = dynamic_cast<const Function&>( t );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  ASSERT( false );
  
  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const StaticTermList& l )
{
  for( size_t i = 0; i < l.GetChildCount(); i++ )
    o << *(l.GetTerm(i));

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const AtomFormula& f )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_ATOMFORMULA );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                    f.GetName() );

  o << dynamic_cast<const StaticTermList&>( f );

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const QuantorFormula& f )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_QUANTIFIEDFORMULA );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_TYPE,
                        f.GetOperatorSymbol()->GetName().c_str() );

  o << *(f.GetVariable());

  o << *(f.GetFormula());

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const UnaryConFormula& f )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_CONJUNCTIVEFORMULA );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_TYPE,
                        f.GetOperatorSymbol()->GetName().c_str() );

  o << *(f.GetFormula());

  o.GetXmlWriter()->CloseElement();
  
  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const BinaryConFormula& f )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_CONJUNCTIVEFORMULA );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_TYPE,
                        f.GetOperatorSymbol()->GetName().c_str() );

  o << *(f.GetLeft());

  o << *(f.GetRight());

  o.GetXmlWriter()->CloseElement();
  
  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const FormulaVariable& f )
{
  o << *(f.GetFormula());

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Formula& f )
{
  try
  {
    const AtomFormula &a = dynamic_cast<const AtomFormula&>( f );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  try
  {
    const QuantorFormula &a = dynamic_cast<const QuantorFormula&>( f );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  try
  {
    const UnaryConFormula &a = dynamic_cast<const UnaryConFormula&>( f );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  try
  {
    const BinaryConFormula &a = dynamic_cast<const BinaryConFormula&>( f );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  try
  {
    const FormulaVariable &a = dynamic_cast<const FormulaVariable&>( f );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  ASSERT( false );

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o,
                                      const FormulaListVariable& l )
{
  if( &l != NULL )
    for( size_t i = 0; i < l.GetChildCount(); i++ )
      o << *(l.GetElement(i));

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o,
                                      const FormulaListElement& e )
{
  try
  {
    const Formula &a = dynamic_cast<const Formula&>( e );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  try
  {
    const FormulaListVariable &a =
      dynamic_cast<const FormulaListVariable&>( e );
    o << a;
    return o;
  }
  catch( bad_cast ) {}

  ASSERT( false );
  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const SElement& e )
{
  try
  {
    const SFormula &a = dynamic_cast<const SFormula&>( e );
    o << *(a.GetFormula());
    return o;
  }
  catch( bad_cast ) {}

  try
  {
    const SFormulaListVariable &a =
      dynamic_cast<const SFormulaListVariable&>( e );
    o << *(a.GetFormulaListVariable());
    return o;
  }
  catch( bad_cast ) {}

  ASSERT( false );
  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const SElementList& l )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_FORMULALIST );

  if( &l != NULL )
    for( size_t i = 0; i < l.GetChildCount(); i++ )
      o << *(l.GetSElement(i));

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Sequent& s )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_SEQUENT );

  o << *(s.GetLeft());

  o << *(s.GetRight());

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Rule& r )
{
  if( r.GetRuleItemCount() == 0 )
  {
    o << *(r.GetConclusion());
    return o;
  }

  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_RULE );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                    r.GetName() );
  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_TYPE,
                                    r.GetType() );

  if( !r.GetParam().empty() )
    o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_PARAM,
                                      r.GetParam() );

  o << *(r.GetConclusion());

  for( size_t i = 0; i < r.GetRuleItemCount(); i++ )
  {
    o << *(r.GetRuleItem(i));
  }

  o.GetXmlWriter()->CloseElement();

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const RuleItem& ruleitem )
{
  try
  {
    const Sequent &s = dynamic_cast<const Sequent&>( ruleitem );
    o << s;
    return o;
  }
  catch( bad_cast )
  {
  }

  try
  {
    const Rule &r = dynamic_cast<const Rule&>( ruleitem );
    o << r;
    return o;
  }
  catch( bad_cast )
  {
  }

  try
  {
    const ProofLink &pl = dynamic_cast<const ProofLink&>( ruleitem );
    Proof *p = pl.GetLinkedProof();
    ASSERT( p != NULL );
    o << *(p->GetEndRule());
    return o;
  }
  catch( bad_cast )
  {
  }

  ASSERT( false ); // unknow type of ruleitem

  return o;
}

////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const Proof& p )
{
  o.GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_PROOF );

  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                    p.GetName() );
  o.GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_CALCULUS,
                                    PROOFDBDTD_DEFAULT_CALCULUS );

  o << *(p.GetEndRule());

  o.GetXmlWriter()->CloseElement();

  return o;
}

#if 0
////////////////////////////////////////////////////////////////////////////////

ExportCeresImpl& operator << ( ExportCeresImpl& o, const ProofDatabase& pdb )
{
  o.Out() << "<proofdatabase>" << endl;

  ProofDatabase::const_iterator i;

  for( i = pdb.begin(); i != pdb.end(); i++ )
  {
    Proof *p = pdb.GetProof( (*i).second->GetName() );

    o <<  *p;
  }

  o.Out() << "</proofdatabase>" << endl;

  return o;
}
#endif
#endif

////////////////////////////////////////////////////////////////////////////////
