////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: term_operations.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream> // TESTING
#include "util/myassert.h"
#include "libproof.h"
#include "term_operations.h"

#include "printhlk.h"

namespace ProofLib
{


////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, TOperationReplaceTerm, v )
{
  if ( np == position )
  {
    return &replacement;
  }

  v->Attach();
  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( Constant, Term, TOperationReplaceTerm, c )
{
  if ( np == position )
  {
    return &replacement;
  }

  c->Attach();
  return c;
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, TOperationReplaceTerm, f )
{
  if ( np == position )
  {
    return &replacement;
  }
  

  TermList *tl = new TermList();
  for ( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i);
    Term *t = DataNode::ApplyTo( f->GetTerm( i ), *this );
    Pop();

    tl->AddRight( t );
  }

  Function *new_f = Function::New( m_ndb, f->GetSymbol(), tl );
  return new_f;
}

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, OperationGetVariables, v )
{
  if( list.FindTerm( v ) < 0 )
    list.AddRight( v );

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( QuantorFormula, Formula, OperationGetVariables, f )
{
  if( list.FindTerm( f->GetVariable() ) < 0 )
    list.AddRight( f->GetVariable() );

  Push(0);
  DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  return f;
}

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, OperationGetSubterms, f )
{
  for ( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Term *t = f->GetTerm( i );

    np.Push( i );
    list.AddRight( t );
    nps.Add( np );
    np.Pop();
  }

  return f;
}

////////////////////////////////////////////////////////////////////////////////
#if 0
DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, OperationGetFirstAndSecondOrderVariables, v )
{
  if( list.FindDefinitionArgument( v ) < 0 )
    list.AddRight( v );

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula, OperationGetFirstAndSecondOrderVariables, f )
{
  SecondOrderVariable *v = f->GetSecondOrderVariable();
  if( v != NULL && list.FindDefinitionArgument( v ) < 0 )
    list.AddRight( v );

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD( QuantorFormula, Formula, OperationGetFirstAndSecondOrderVariables, f )
{
  if( list.FindDefinitionArgument( f->GetVariable() ) < 0 )
    list.AddRight( f->GetVariable() );

  Push(0);
  DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD( SecondOrderQuantorFormula, Formula, OperationGetFirstAndSecondOrderVariables, f )
{
  if( list.FindDefinitionArgument( f->GetSecondOrderVariable() ) < 0 )
    list.AddRight( f->GetSecondOrderVariable() );

  Push(0);
  DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  return f;
}
#endif
////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD
  ( Variable, Term, OperationCountFreeVariables, v )
{
  if( bounded.FindTerm( v ) >= 0 )
    return v;

  varcount++;

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationCountFreeVariables, f )
{
  if( bounded.FindTerm( f->GetVariable() ) < 0 )
    bounded.AddRight( f->GetVariable() );

  Push(1);
  DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  int pos = bounded.FindTerm( f->GetVariable() );

  if( pos >= 0 )
    bounded.RemoveAt( pos );

  return f;
}

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD
  ( Variable, Term, OperationGetFreeVariablesAndPositions, v )
{
  if( bounded.FindTerm( v ) >= 0 )
    return v;

  if( list.FindTerm( v ) < 0 )
  {
    list.AddRight( v );
    size_t p = pos.size();
    pos.resize( p + 1 );
    pos[p].Add( np );
  }

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationGetFreeVariablesAndPositions, f )
{
  if( bounded.FindTerm( f->GetVariable() ) < 0 )
    bounded.AddRight( f->GetVariable() );

  Push(1);
  DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  int pos = bounded.FindTerm( f->GetVariable() );

  if( pos >= 0 )
    bounded.RemoveAt( pos );

  return f;
}

const NodePointers&
OperationGetFreeVariablesAndPositions::GetPositions( size_t index ) const
{
  ASSERT( index < list.GetChildCount() );
  return pos[index];
}

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationGetFreeVariables, f )
{
  bool bMyVarIsAlsoFree = ( list.FindTerm( f->GetVariable() ) >= 0 );

  FormulaOperation::ApplyToQuantorFormula( f );

  if( bMyVarIsAlsoFree )
    return f;

  int pos = list.FindTerm( f->GetVariable() );

  if( pos >= 0 )
    list.RemoveAt( pos );

  return f;
}

////////////////////////////////////////////////////////////////////////////////
#if 0
DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationGetFreeFirstAndSecondOrderVariables, f )
{
  bool bMyVarIsAlsoFree = ( list.FindDefinitionArgument( f->GetVariable() ) >= 0 );

  FormulaOperation::ApplyToQuantorFormula( f );

  if( bMyVarIsAlsoFree )
    return f;

  int pos = list.FindDefinitionArgument( f->GetVariable() );

  if( pos >= 0 )
    list.RemoveAt( pos );

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SecondOrderQuantorFormula, Formula, OperationGetFreeFirstAndSecondOrderVariables, f )
{
  bool bMyVarIsAlsoFree = ( list.FindDefinitionArgument( f->GetSecondOrderVariable() ) >= 0 );

  FormulaOperation::ApplyToSecondOrderQuantorFormula( f );

  if( bMyVarIsAlsoFree )
    return f;

  int pos = list.FindDefinitionArgument( f->GetSecondOrderVariable() );

  if( pos >= 0 )
    list.RemoveAt( pos );

  return f;
}
#endif

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationGetBoundedVariables, f )
{
  if( list.FindTerm( f->GetVariable() ) < 0 )
    list.AddRight( f->GetVariable() );

  return f;
}

////////////////////////////////////////////////////////////////////////////////

void OperationGetVariableSubstitutions::Clear()
{
  while( vars.GetChildCount() > 0 ) vars.RemoveAt(0);
  while( reps.GetChildCount() > 0 ) reps.RemoveAt(0);
  while( bounded.GetChildCount() > 0 ) bounded.RemoveAt(0);
  while( equals.GetChildCount() > 0 ) equals.RemoveAt(0);
  pos.clear();
  np.Clear();
}

bool OperationGetVariableSubstitutions::CheckForSubstitutions(
  DataNode *general, DataNode *instance )
{
  Clear();
  return general->ApplyCompare( *this, instance );
}

const Variable* OperationGetVariableSubstitutions::GetVar( size_t index ) const
{
  ASSERT( index < GetCount() );
  return vars.GetTerm(index)->CastTo<const Variable>();
}

const Term* OperationGetVariableSubstitutions::GetSubst( size_t index ) const
{
  ASSERT( index < GetCount() );
  return reps.GetTerm(index);
}

const NodePointer& OperationGetVariableSubstitutions::GetVarPos( size_t index ) const
{
  ASSERT( index < GetCount() );
  return pos[index].var;
}

const NodePointers&
OperationGetVariableSubstitutions::GetSubstPos( size_t index ) const
{
  ASSERT( index < GetCount() );
  return pos[index].substs;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Variable, OperationGetVariableSubstitutions, v )
{
  if( !equals_too && v == o )
  {
    // this var is already substituted, partial substition not allowed here!
    if( vars.FindTerm( v ) >= 0 )
      return false;

    equals.AddRight( (Term*) v );

    return true;
  }

  const Term *t = o->CastTo<const Term>();

  if( t == NULL )
    return false;

  // variable already found?
  int p = vars.FindTerm( v );
  if( p >= 0 )
  {
    // variable has to be substituted with the same term!
    if( reps.GetTerm(p) != t )
      return false;

    pos[p].substs.Add( np );
    return true;
  }

  // was there already a position, where this var wasn't substituted?
  if( equals.FindTerm( v ) >= 0 ) 
    return false;

  // is this a bounded variable?
  if( bounded.FindTerm( v ) >= 0 ) 
    return true;

  p = pos.size();
  pos.resize( p + 1 );
  pos[p].var = np;
  pos[p].substs.Add( np );
  vars.AddRight( (Term*) v );
  reps.AddRight( (Term*) t );

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( QuantorFormula, OperationGetVariableSubstitutions, f )
{
  Variable *v = f->GetVariable();

  bool already = ( bounded.FindTerm( v ) >= 0 );

  if( !already )
    bounded.AddRight( v );

  bool ok = FormulaCompareOperation::ApplyToQuantorFormula( f, o );

  if( !already )
  {
    int pos = bounded.FindTerm( v );
    bounded.RemoveAt( pos );
  }

  return ok;
}

////////////////////////////////////////////////////////////////////////////////

OperationVariableSubstitution::OperationVariableSubstitution(
  ProofDatabase *pdb, const Variable *v, const Term *subst )
{
  ASSERT( pdb != NULL );
  m_ndb = pdb->GetNodeDatabase();
  m_pdb = pdb;

  AddSubstitution( v, subst );
}

OperationVariableSubstitution::OperationVariableSubstitution(
  NodeDatabase *ndb, const Variable *v, const Term *subst )
{
  ASSERT( ndb != NULL );
  m_ndb = ndb;
  m_pdb = NULL;

  AddSubstitution( v, subst );
}

OperationVariableSubstitution::OperationVariableSubstitution(
  NodeDatabase *ndb, const VariableSubstitution &subst )
{
  ASSERT( ndb != NULL );
  m_ndb = ndb;
  m_pdb = NULL;

  for( size_t i = 0; i < subst.GetCount(); i++ )
    AddSubstitution( subst.GetVar(i), subst.GetSubst(i) );
}

OperationVariableSubstitution::OperationVariableSubstitution(
  NodeDatabase *ndb )
{
  ASSERT( ndb != NULL );
  m_ndb = ndb;
  m_pdb = NULL;
}

OperationVariableSubstitution::OperationVariableSubstitution(
  ProofDatabase *pdb )
{
  ASSERT( pdb != NULL );
  m_pdb = pdb;
  m_ndb = pdb->GetNodeDatabase() ;
}

void OperationVariableSubstitution::AddSubstitution(
  const Variable *v, const Term *subst )
{
  ASSERT( v != NULL );
  ASSERT( subst != NULL );
  substitutions.push_back( Substitution( v, subst ) );
  m_positions.resize( substitutions.size() );
}

const Variable* OperationVariableSubstitution::GetVar( size_t index ) const
{
  ASSERT( index < GetCount() );
  return substitutions[index].first;
}

const Term* OperationVariableSubstitution::GetSubst( size_t index ) const
{
  ASSERT( index < GetCount() );
  return substitutions[index].second;
}

const NodePointers&
OperationVariableSubstitution::GetPositions( size_t index ) const
{
  ASSERT( index < substitutions.size() );
  ASSERT( index < m_positions.size() );

  return m_positions[index];
}

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, OperationVariableSubstitution, v )
{
#if 0
  if( v != this->v )
    return v;

  return (Term*) t;
#endif
  for( size_t i = 0; i < substitutions.size(); i++ )
  {
    if( v != substitutions[i].first )
      continue; // not "this" variable

    if( bounded.FindTerm( substitutions[i].first ) >= 0 )
      continue; // variable is bounded

    m_positions[i].Add( m_np );
    return (Term*) substitutions[i].second;
  }

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationVariableSubstitution, f )
{
  // don't replace bounded vars!
  bounded.AddRight( f->GetVariable() );
#if 0
  if( v == f->GetVariable() )
   return f;
#endif

  ProofLib::Formula *f_1 = FormulaOperation::ApplyToQuantorFormula( f );

  ASSERT( bounded.GetTerm( bounded.GetChildCount() - 1 ) == f->GetVariable() );
  bounded.RemoveAt( bounded.GetChildCount() - 1 );

  return f_1;
}

std::ostream& OperationVariableSubstitution::ToString( std::ostream &out ) const
{
  typedef std::pair< const Variable*, const Term* > Substitution;

  for( size_t i = 0; i < substitutions.size(); i++ )
  {
    if( i != 0 ) out << ", ";
    substitutions[i].first->ToString(out);
    out << " <- ";
    substitutions[i].second->ToString(out);
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

OperationTermSubstitution::OperationTermSubstitution(
  NodeDatabase *ndb, const Term *t, const Term *subst,
  const NodePointers* positions_allowed )
{
  ASSERT( ndb != NULL );
//  m_ndb = pdb->GetNodeDatabase();
//  m_pdb = pdb;
  m_ndb = ndb;

  this->t = t;
  this->subst = subst;

  m_positions_allowed = positions_allowed;
}

void OperationTermSubstitution::Push( size_t index )
{
  m_np.Push( index );
}

size_t OperationTermSubstitution::Pop()
{
  return m_np.Pop();
}

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, OperationTermSubstitution, v )
{
  if( m_positions_allowed != NULL && !m_positions_allowed->IsAboveAny( m_np ) )
    return v;

  if( v == t )
  {
    m_positions.Add( m_np );
    return (Term*) subst;
  }

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( Constant, Term, OperationTermSubstitution, c )
{
  if( m_positions_allowed != NULL && !m_positions_allowed->IsAboveAny( m_np ) )
    return c;

  if( c == t )
  {
    m_positions.Add( m_np );
    return (Term*) subst;
  }

  return c;
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, OperationTermSubstitution, f )
{
  if( m_positions_allowed != NULL && !m_positions_allowed->IsAboveAny( m_np ) )
    return f;

  if( f == t )
  {
    m_positions.Add( m_np );
    return (Term*) subst;
  }

  return TermOperation::ApplyToFunction( f );
}

////////////////////////////////////////////////////////////////////////////////

const Term* VariableSubstitution::GetSubstFor( const Variable *v ) const
{
  for( size_t i = 0; i < GetCount(); i++ )
  {
    if( GetVar(i) == v )
      return GetSubst(i);
  }

  return NULL;
}

////////////////////////////////////////////////////////////////////////////////

bool OperationMGUSimple::Unify( const DataNode *left, const DataNode *right )
{
  return left->ApplyCompare( *this, right );
}

std::ostream& OperationMGUSimple::ToString( std::ostream &out ) const
{
  UnificationTable::const_iterator i;

  for( i = m_table.begin(); i != m_table.end(); i++ )
  {
    if( i != m_table.begin() )
      out << ",";

    i->first->ToString(out);
    out << " == ";
    i->second->ToString(out);
  }

  return out;
}

bool OperationMGUSimple::AddSubstition( Side s, const Variable *v, const Term *t )
{
  ASSERT( v != NULL );
  ASSERT( t != NULL );

  const Term* subst = GetSubstFor(v);

  OperationVariableSubstitution op( GetNodeDatabase(), *this );

  Term *t1 = DataNode::ApplyTo( (Term*) t, op );

  if( subst == t1 )
    return true;

  TermList check;
  OperationGetVariables op_oc( check );

  DataNode::ApplyTo( (Term*) t1, op_oc );

  if( check.FindTerm( v ) >= 0 )
    return false; // occur check!

  if( subst == NULL )
  {
    OperationVariableSubstitution op1( GetNodeDatabase() );
    op1.AddSubstitution( v, t1 );
    UnificationTable::iterator i;
    for( i = m_table.begin(); i != m_table.end(); i++ )
    {
      i->second = DataNode::ApplyTo( (Term*) i->second, op1 );
    }
    m_table.push_back( Unification( v, t1 ) );
    return true;
  }

  const Variable *subst_v = subst->CastTo<const Variable>();

  if( subst_v != NULL )
  {
    if( check.FindTerm( subst_v ) >= 0 )
      return false; // occur check!

    OperationVariableSubstitution op1( GetNodeDatabase() );
    op1.AddSubstitution( v, t1 );
    op1.AddSubstitution( subst_v, t1 );
    UnificationTable::iterator i;
    for( i = m_table.begin(); i != m_table.end(); i++ )
    {
      i->second = DataNode::ApplyTo( (Term*) i->second, op1 );
    }
    m_table.push_back( Unification( subst_v, t1 ) );
    return true;
  }

  return false;
}

const Variable* OperationMGUSimple::GetVar( size_t index ) const
{
  ASSERT( index < m_table.size() );
  return m_table[index].first;
}

const Term* OperationMGUSimple::GetSubst( size_t index ) const
{
  ASSERT( index < m_table.size() );
  return m_table[index].second;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Variable, OperationMGUSimple, t )
{
  const Term *o_t = o->CastTo<const Term>();
  return AddSubstition( Left, t, o_t );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Constant, OperationMGUSimple, t )
{
  if( t == o )
    return true;

  const Variable *o_v = o->CastTo<const Variable>();
  
  if( o_v != NULL )
    return AddSubstition( Right, o_v, t );

  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Function, OperationMGUSimple, t )
{
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  const Function *f_o = o->CastTo<const Function>();

  if( f_o == NULL )
  {
    const Variable *o_v = o->CastTo<const Variable>();

    if( o_v != NULL )
      return AddSubstition( Right, o_v, t );

    return false;
  }

  if( !t->HaveEqualSymbol( f_o ) )
    return false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////

FreshVarMap::FreshVarMap( FreshVariableGenerator *fg, const DataNode *node,
                          size_t start )
{
  ASSERT( fg != NULL );
  ASSERT( node != NULL );
  m_fg = fg;
  m_next_var_num = start;
  AddVarsFromNode( node );
}

FreshVarMap::~FreshVarMap()
{
  for( VarMapping::iterator i = m_map.begin(); i != m_map.end(); i++ )
  {
    i->first->Detach();
    i->second->Detach();
  }
}

void FreshVarMap::operator=( const FreshVarMap &o )
{
  for( VarMapping::iterator i = m_map.begin(); i != m_map.end(); i++ )
  {
    i->first->Detach();
    i->second->Detach();
  }

  m_map.clear();

  for( VarMapping::const_iterator i = o.m_map.begin(); i != o.m_map.end(); i++ )
  {
    i->first->Attach();
    i->second->Attach();

    m_map.push_back( VarMap( i->first, i->second ) );
  }

  m_fg = o.m_fg;
  m_next_var_num = o.m_next_var_num;
}

void FreshVarMap::AddVarsFromNode( const DataNode *node )
{
  ASSERT( node != NULL );

  TermList vars;
  OperationGetVariables getvars( vars );

  ((DataNode*) node)->Apply( getvars );
  
  for( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    Variable* v = vars.GetTerm(i)->CastTo<Variable>();
    v->Attach();
    Variable* w = m_fg->GetVariableNum( m_next_var_num++ );
    w->Attach();
    m_map.push_back( VarMap( v, w ) );
  }
}

const Variable* FreshVarMap::GetVar( size_t index ) const
{
  ASSERT( index < m_map.size() );
  return m_map[index].first;
}

const Variable* FreshVarMap::GetSubst( size_t index ) const
{
  ASSERT( index < m_map.size() );
  return m_map[index].second;
}

////////////////////////////////////////////////////////////////////////////////

OperationMGU::UnificationTable::~UnificationTable()
{
  for( size_t i = 0; i < size(); i++ )
  {
    (*this)[i].first->Detach();
    (*this)[i].second->Detach();
  }
}

void OperationMGU::UnificationTable::operator=( const UnificationTable &o )
{
  for( size_t i = 0; i < size(); i++ )
  {
    (*this)[i].first->Detach();
    (*this)[i].second->Detach();
  }

  clear();

  for( size_t i = 0; i < o.size(); i++ )
  {
    AddSubstitution( o[i].first, o[i].second );
  }
}

void OperationMGU::UnificationTable::
AddSubstitution( const Variable* v_, const Term* t_ )
{
  Variable *v = (Variable*) v_;
  Term *t = (Term*) t_;
  
  v->Attach();
  t->Attach();

  push_back( Unification( v, t ) );
}

void OperationMGU::UnificationTable::ApplyOperation( TermOperation &op )
{
  UnificationTable::iterator i;
  for( i = begin(); i != end(); i++ )
  {
    Term *t = i->second;
    i->second = DataNode::ApplyTo( i->second, op );

    if( t != i->second )
    {
      i->second->Attach();
      t->Detach();
    }
  }
}

const Variable* OperationMGU::UnificationTable::GetVar( size_t index ) const
{
  ASSERT( index < GetCount() );
  return (*this)[index].first;
}

const Term* OperationMGU::UnificationTable::GetSubst( size_t index ) const
{
  ASSERT( index < GetCount() );
  return (*this)[index].second;
}

std::ostream&
OperationMGU::UnificationTable::ToString( std::ostream &out ) const
{
  const_iterator i;

  for( i = begin(); i != end(); i++ )
  {
    if( i != begin() )
      out << ",";

    i->first->ToString(out);
    out << " == ";
    i->second->ToString(out);
  }

  return out;
}

OperationMGU::OperationMGU( FreshVariableGenerator *fg,
                            const FreshVarMap *m1, const FreshVarMap *m2 )
{
  m_fg = fg;
  m_new_var_count = 0;
  m_map_l = m1;
  m_map_r = m2;
}

/*
void OperationMGU::operator=( const OperationMGU &o )
{
   m_tabla = o.m_tablee;

  const FreshVarMap *m_map_l;
  const FreshVarMap *m_map_r;

  size_t m_new_var_count;

  FreshVariableGenerator *m_fg;
}
*/

bool OperationMGU::Unify( const DataNode *left, const DataNode *right )
{
  return left->ApplyCompare( *this, right );
}

std::ostream& OperationMGU::ToString( std::ostream &out ) const
{
#if 0
  UnificationTable::const_iterator i;

  for( i = m_table.begin(); i != m_table.end(); i++ )
  {
    if( i != m_table.begin() )
      out << ",";

    i->first->ToString(out);
    out << " == ";
    i->second->ToString(out);
  }
#endif
//  m_table.ToString(out);

  bool left = true;

  while( true )
  {
    out << ( left ? "left" : " right" ) << ": ";
    
    const FreshVarMap *map = ( left ? m_map_l : m_map_r );
    
    for( size_t i = 0; i < map->GetCount(); i++ )
    {
      if( i != 0 ) out << ", ";
      map->GetVar(i)->ToString(out);
      out << " <- ";
      const Term *t = m_table.GetSubstFor( map->GetSubst(i) );
      if( t == NULL )
        map->GetSubst(i)->ToString(out);
      else
        t->ToString(out);
    }

    if( left ) left = false; else break;
  }

  return out;
}

void
OperationMGU::GetSubstitution( Side s, OperationVariableSubstitution &op ) const
{
  const FreshVarMap *map = ( s == Left ? m_map_l : m_map_r );

  for( size_t i = 0; i < map->GetCount(); i++ )
  {
    const Term *t = m_table.GetSubstFor( map->GetSubst(i) );
    if( t == NULL ) t = map->GetSubst(i);
    op.AddSubstitution( map->GetVar(i), t );
  }
}

const Variable* OperationMGU::MapVar( Side s, const Variable *v )
{
  ASSERT( v != NULL );

  const FreshVarMap *map = ( s == Left ? m_map_l : m_map_r );

#if 1
  if( map->GetSubstFor(v) == NULL )
  {
    std::cerr << "Missing map on " << ( s == Left ? "left" : "right" )
              << " side for var ";
    v->ToString(std::cerr) << "." << std::endl;
  }
#endif
  return map->GetSubstFor(v)->CastTo<const Variable>();
#if 0
  for( size_t i = 0; i < map.size(); i++ )
  {
    if( map[i].first == v )
      return map[i].second;
  }
  
  Variable* w = m_fg->GetVariableNum( m_new_var_count++ );
  w->Attach(); // TODO Detach sometime !!!
  ASSERT( w != NULL );
  map.push_back( VarMap( v, w ) );
  return w;
#endif
}

bool OperationMGU::AddSubstition( Side s, const Variable *v, const Term *t )
{
  ASSERT( v != NULL );
  ASSERT( t != NULL );

  v = MapVar( s, v );

  TermList list;
  OperationGetVariables getVars( list );
  ((Term*) t)->Apply( getVars );

  OperationVariableSubstitution vs( GetNodeDatabase() );

  for( size_t i = 0; i < list.GetChildCount(); i++ )
  {
    Variable *u = list.GetTerm(i)->CastTo<Variable>();
    ASSERT( u != NULL );

    vs.AddSubstitution( u, MapVar( ( s == Left ? Right : Left ), u ) );
  }

  Term *u = DataNode::ApplyTo( (Term*) t, vs );
  u->Attach();
  
  bool ok = AddSubstition( v, u );
  
  u->Detach();

  return ok;
}

bool OperationMGU::AddSubstition( const Variable *v, const Term *t )
{
  ASSERT( v != NULL );
  ASSERT( t != NULL );

  const Term* subst = m_table.GetSubstFor(v);

  OperationVariableSubstitution op( GetNodeDatabase(), m_table );

  Term *t1 = DataNode::ApplyTo( (Term*) t, op );

  if( subst == t1 )
    return true;

  TermList check;
  OperationGetVariables op_oc( check );

  DataNode::ApplyTo( (Term*) t1, op_oc );

  if( check.FindTerm( v ) >= 0 )
    return false; // occur check!

  if( subst == NULL )
  {
    OperationVariableSubstitution op1( GetNodeDatabase() );
    op1.AddSubstitution( v, t1 );
    m_table.ApplyOperation( op1 );
    m_table.AddSubstitution( v, t1 );
#if 0
    UnificationTable::iterator i;
    for( i = m_table.begin(); i != m_table.end(); i++ )
    {
      i->second = DataNode::ApplyTo( (Term*) i->second, op1 );
    }
    m_table.push_back( Unification( v, t1 ) );
#endif
    return true;
  }

  const Variable *subst_v = subst->CastTo<const Variable>();

  if( subst_v != NULL )
  {
    if( check.FindTerm( subst_v ) >= 0 )
      return false; // occur check!

    OperationVariableSubstitution op1( GetNodeDatabase() );
    op1.AddSubstitution( v, t1 );
    op1.AddSubstitution( subst_v, t1 );
    m_table.ApplyOperation( op1 );
    m_table.AddSubstitution( subst_v, t1 );
#if 0
    UnificationTable::iterator i;
    for( i = m_table.begin(); i != m_table.end(); i++ )
    {
      i->second = DataNode::ApplyTo( (Term*) i->second, op1 );
    }
    m_table.push_back( Unification( subst_v, t1 ) );
#endif
    return true;
  }

  return false;
}

void OperationMGU::Finalize()
{
  m_map_l = new FreshVarMap( *m_map_l );
  m_map_r = new FreshVarMap( *m_map_r );
}

#if 0
const Variable* OperationMGU::GetVar( size_t index ) const
{
  ASSERT( index < m_table.GetCount() );
  return m_table.GetVar(index);
}

const Term* OperationMGU::GetSubst( size_t index ) const
{
  ASSERT( index < m_table.GetCount() );
  return m_table.GetSubst(index);
}
#endif

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Variable, OperationMGU, t )
{
  const Term *o_t = o->CastTo<const Term>();
  return AddSubstition( Left, t, o_t );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Constant, OperationMGU, t )
{
  if( t == o )
    return true;

  const Variable *o_v = o->CastTo<const Variable>();
  
  if( o_v != NULL )
    return AddSubstition( Right, o_v, t );

  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Function, OperationMGU, t )
{
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  const Variable *o_v = o->CastTo<const Variable>();

  if( o_v != NULL )
    return AddSubstition( Right, o_v, t );

#if 0
  if( !t->HaveEqualSymbol( f_o ) )
    return false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }
#endif

  return TermCompareOperation::ApplyToFunction( t, o );
}

////////////////////////////////////////////////////////////////////////////////

Substitution::Substitution()
{
  var = NULL;
  by = NULL;
}

Substitution::Substitution( const Variable *v, const Term *r )
{
  ASSERT( v != NULL );
  ASSERT( r != NULL );

  var = (Variable*) v;
  by = (Term*) r;

  var->Attach();
  by->Attach();
}

Substitution::Substitution( const Substitution &o )
{
  (*this) = o;
}

Substitution::~Substitution()
{
  if( var != NULL )
    var->Detach();
  if( by != NULL )
    by->Detach();
}

void Substitution::operator=( const Substitution &o )
{
  var = o.var;
  by = o.by;

  if( var != NULL )
    var->Attach();
  if( by != NULL )
    by->Attach();
}

std::ostream& Substitution::ToString( std::ostream &out ) const
{
  if( var == NULL )
  {
    out << "(empty substitution)";
    return out;
  }

  ASSERT( by != NULL );

  var->ToString(out);
  out << " <-- ";
  by->ToString(out);

  return out;
}

////////////////////////////////////////////////////////////////////////////////

const Term* Substitutions::GetReplacementFor( const Variable *var ) const
{
  for( size_t i = 0; i < size(); i++ )
  {
    if( (*this)[i].GetVar() == var )
      return (*this)[i].GetSubst();
  }

  return NULL;
}

bool Substitutions::CheckSubstitution( const Variable *var, const Term *rep )
{
  const Term *ex = GetReplacementFor( var );

  if( ex == NULL )
  {
    push_back( Substitution( var, rep ) );
    return true;
  }

  if( ex == rep )
    return true;

  return false;
}

// subset of s?
bool Substitutions::operator < ( const Substitutions &s ) const
{
  if( size() > s.size() )
    return false;

  for( size_t i = 0; i < size(); i++ )
    if( (*this)[i].GetSubst() != s.GetReplacementFor( (*this)[i].GetVar() ) )
      return false;

  return true;
}

std::ostream& Substitutions::ToString( std::ostream &out ) const
{
  for( size_t i = 0; i < size(); i++ )
  {
    (*this)[i].ToString( out );
    out << std::endl;
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

void SubstPossibilities::Add( const Substitutions &s )
{
  for( size_t i = 0; i < size(); i++ )
  {
    if( s < (*this)[i] )
      return; // s is a subset of an already existing subst
    if( (*this)[i] < s )
    {
      (*this)[i] = s; // s is a super set of an ex subst.
      return;
    }
  }

  push_back( s );
}

bool SubstPossibilities::Merge( const SubstPossibilities &o )
{
  SubstPossibilities ex( *this );

  bool was_empty = empty();

  clear();

  for( size_t i = 0; i < o.size(); i++ )
  {
    if( ex.empty() )
    {
      push_back( o[i] );
      continue;
    }

    for( size_t j = 0; j < ex.size(); j++ )
    {
      Substitutions s( ex[j] );

      bool failed = false;

      for( size_t k = 0; k < o[i].size(); k++ )
      {
        if( !s.CheckSubstitution( o[i][k].GetVar(), o[i][k].GetSubst() ) )
        {
          failed = true;
          break;
        }
      }

      if( !failed )
      {
        push_back( s );
      }
    }
  }
  
  return was_empty || !empty();
}

std::ostream& SubstPossibilities::ToString( std::ostream &out ) const
{
  for( size_t i = 0; i < size(); i++ )
  {
    out << i << "." << std::endl;
    (*this)[i].ToString( out );
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

}
