////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: transhlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>

#include "libproof/util/myassert.h"
#include "libproof/util/exceptions.h"
#include "libproof/util/hash.h"
#include "libproof/chars/latexchars.h"
#include "libproof/libproof.h"
#include "libproof/operations.h"
#include "libproof/term_operations.h"

#include "cereslk.h"

#include "transhlk.h"
#include "symbolhlk.h"
#include "printhlk.h"
#include "typehlk.h"

#ifdef DEBUG
#define DEBUG_PRINT_INFO
#endif

////////////////////////////////////////////////////////////////////////////////

namespace ProofLib
{

class CheckInstanceOperation : public TermCompareOperation
{
protected:
  Substitutions current;
  SubstPossibilities possibilities;

public:
  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  CheckInstanceOperation() {}
  virtual ~CheckInstanceOperation() {}

  bool IsInstanceOf( Sequent *instance, Sequent *schema )
  {
    Clear();
    return schema->ApplyCompare( *this, instance );
  }

  void Clear()
  {
    ClearCurrent();
    possibilities.clear();
  }

  size_t GetCount() const { return possibilities.size(); }

protected:
  void ClearCurrent()
  {
    current.clear();
  }

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SElementList );
};

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Variable, CheckInstanceOperation, t )
{
  const Term *rep = o->CastTo<const Term>();

  return current.CheckSubstitution( t, rep );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SElementList, CheckInstanceOperation, l )
{
  if( o == NULL )
    return l == NULL;

  const SElementList *l_o = o->CastTo<const SElementList>();

  ASSERT( l_o != NULL );

  if( l == NULL )
    return false;
  
#if 0
  bool ok = true;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    SubstPossibilities p;

    SFormula *sf_i = l->GetSElement(i)->CastTo<SFormula>();
    ASSERT( sf_i != NULL );

    bool found = false;

    for( size_t j = 0; j < l_o->GetChildCount(); j++ )
    {
      SFormula *sf_j = l_o->GetSElement(j)->CastTo<SFormula>();
      ASSERT( sf_j != NULL );

      if( sf_i->ApplyCompare( *this, sf_j ) )
      {
        found = true;
        break;
      }
    }

    if( !found )
    {
      ok = false;
      break;
    }
  }

  return ok;
#endif
  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    SubstPossibilities p;

    for( size_t j = 0; j < l_o->GetChildCount(); j++ )
    {
      ClearCurrent();

      Push(i);
      bool ok = l->GetSElement(i)->ApplyCompare( *this, l_o->GetSElement(j) );
      Pop();
      
      if( ok )
        p.Add( current );
    }

    if( p.empty() )
      return false;

    if( !possibilities.Merge(p) )
      return false;
  }

  return true;
}

#if 0 // OLD VERSION, incorrect (does not detect all intances)
/// checks an construct if its an instance of the other
class CheckInstanceOperation : public TermCompareOperation
{
  stlhash::hash_map< const DbNode*, const DataNode* > m_map;
/*
protected:
  TermList terms;
  TermList reps;

  typedef std::vector<NodePointer> Occurrences;
  std::vector<Occurrences> m_nps;

  NodePointer m_np;

public:
  virtual void Push( size_t index ) { m_np.Push( index ); }
  virtual size_t Pop() { return m_np.Pop(); }
*/
public:
  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  CheckInstanceOperation() {}
  virtual ~CheckInstanceOperation() {}

  void Clear()
  {
/*
    while( terms.GetChildCount() > 0 ) terms.RemoveAt(0);
    while( reps.GetChildCount() > 0 ) reps.RemoveAt(0);
    m_nps.clear();
    m_np.Clear();
*/
  }

  bool IsInstanceOf( Sequent *instance, Sequent *schema )
  {
    Clear();
    bool ok = schema->ApplyCompare( *this, instance );

    return ok;
  }

  size_t GetCount() const { return m_map.size(); }

/*
  size_t GetCount() const { return terms.GetChildCount(); }
  const Term* GetTerm( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return terms.GetTerm(index);
  }

  const Term* GetSubst( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return reps.GetTerm(index);
  }

  const Occurrences& GetOccurrences( size_t index )
  {
    ASSERT( index < GetCount() );
    return m_nps[index];
  }
*/

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SElementList );

protected:
/*
  bool CheckTerms( Term *l, DataNode *r_ )
  {
    if( l == r_ )
      return true;

    Term *r = r_->CastTo<Term>();
  
    if( r == NULL )
      return false;
  
    int pos = terms.FindTerm( l );

    if( pos >= 0 && GetSubst( pos ) == r )
    {
      m_nps[pos].push_back( m_np );
      return true;
    }

#if 0
HLK::Print::PrintTerm( l, std::cerr );
std::cerr << " <- ";
HLK::Print::PrintTerm( r, std::cerr );
std::cerr << std::endl;
#endif

    terms.AddRight( (Term*) l );
    reps.AddRight( (Term*) r );
    m_nps.resize( m_nps.size() + 1 );
    m_nps.back().push_back( m_np );

    return true;
  }
*/
};

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Variable, CheckInstanceOperation, t )
{
  if( m_map.find( t ) == m_map.end() )
  {
    m_map[t] = o;
    return true;
  }

  return m_map[t] == o;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Constant, CheckInstanceOperation, t )
{
  return t == o;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Function, CheckInstanceOperation, t )
{
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  if( t == o )
    return true;

  const Function *f_o = o->CastTo<const Function>();

  if( f_o == NULL || t->GetSymbol() != f_o->GetSymbol() )
    return false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SElementList, CheckInstanceOperation, l )
{
  const SElementList *l_o = o->CastTo<const SElementList>();

  if( l_o == NULL )
    return l == NULL;

  if( l == NULL )
    return false;
  
  bool ok = true;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    SFormula *sf_i = l->GetSElement(i)->CastTo<SFormula>();
    ASSERT( sf_i != NULL );

    bool found = false;

    for( size_t j = 0; j < l_o->GetChildCount(); j++ )
    {
      SFormula *sf_j = l_o->GetSElement(j)->CastTo<SFormula>();
      ASSERT( sf_j != NULL );

      if( sf_i->ApplyCompare( *this, sf_j ) )
      {
        found = true;
        break;
      }
    }

    if( !found )
    {
      ok = false;
      break;
    }
  }

  return ok;
}
#endif

}

////////////////////////////////////////////////////////////////////////////////

namespace HLK
{

class TransformerImpl : public Transformer
{
  ProofLib::Proof         *m_lk_root_proof;
  ProofLib::Proof         *m_lk_proof;
  ProofLib::ProofDatabase *m_lk_pdb;

  AxiomTable              *m_axiom_table;

  typedef stlhash::hash_set<ProofLib::Formula*> FormulaSet;

private:
  TransformerImpl( const TransformerImpl& ) {};

public:
  TransformerImpl( AxiomTable* axiomtbl )
  {
    m_lk_root_proof = NULL;
    m_lk_proof = NULL;
    m_lk_pdb = NULL;
    m_axiom_table = axiomtbl;
  }

  virtual ~TransformerImpl() {}

public:
  AxiomTable* GetAxiomTable() { return m_axiom_table; }

  virtual void CompleteProof( ProofLib::Proof* p,
                              ProofLib::ProofDatabase* pdb )
  {
    ASSERT( p != NULL );
    ASSERT( pdb != NULL );

    m_lk_root_proof = p;
    m_lk_proof = p;
    m_lk_pdb = pdb;

#if 0
    std::cerr << "Translate proof: " << p->GetName() << std::endl;
#endif

    FormulaSet world_l, world_r;

    ProofLib::Rule *rule = p->GetEndRule()->CastTo<ProofLib::Rule>();
    ASSERT( rule != NULL );
    // build world
    AddSequentToWorld( rule->GetConclusion(), world_l, world_r );

    // start rekursive proof completion
    ProofLib::Rule *start_rule = rule->GetRuleItem(0)->CastTo<ProofLib::Rule>();

    bool remove_help_rule = false;
    ProofLib::Rule *help_rule = NULL;
    if( start_rule == NULL )
    {
      ProofLib::ProofLink *pl =
        rule->GetRuleItem(0)->CastTo<ProofLib::ProofLink>();

      ASSERT( pl != NULL );

/*
      return CompleteProof( pl->GetLinkedProof(), pdb );
ASSERT( false );
*/

      //start_rule = rule;
      start_rule = new ProofLib::Rule;
      rule->ReplaceRuleItem( pl, start_rule );
      start_rule->SetConclusion( rule->GetConclusion()->Clone() );
      start_rule->AddRight( pl );
      remove_help_rule = true;
      help_rule = start_rule;
    }
#ifdef DEBUG_PRINT_INFO
std::cerr << " BEFORE ------------------------------------------------------------------------------" << std::endl;
Print::PrintSequent( start_rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
Print::PrintSequent( rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
std::cerr << " ------------------------------------------------------------------------------" << std::endl;
#endif

    start_rule = CompleteProof( start_rule, world_l, world_r );

#ifdef DEBUG_PRINT_INFO
std::cerr << " AFTER ------------------------------------------------------------------------------" << std::endl;
Print::PrintSequent( start_rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
Print::PrintSequent( rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
std::cerr << " ------------------------------------------------------------------------------" << std::endl;
#endif
    PermuationRuleBuilder perm_builder( start_rule, true, m_lk_proof );
    perm_builder.OrderEqualsTogether();
    start_rule = perm_builder.BuildRule();

    PermuationRuleBuilder perm_builder_r( start_rule, false, m_lk_proof );
    perm_builder_r.OrderEqualsTogether();
    start_rule = perm_builder_r.BuildRule();

    ContractionRuleBuilder cont_builder( start_rule, true, m_lk_proof );
    cont_builder.ContractTo( rule->GetConclusion() );
    start_rule = cont_builder.BuildRule();

    ContractionRuleBuilder cont_builder_r( start_rule, false, m_lk_proof );
    cont_builder_r.ContractTo( rule->GetConclusion() );
    start_rule = cont_builder_r.BuildRule();

    WeakeningRuleBuilder weak_builder( start_rule, true, m_lk_proof );
    weak_builder.WeakTo( rule->GetConclusion() );
    start_rule = weak_builder.BuildRule();

    WeakeningRuleBuilder weak_builder_r( start_rule, false, m_lk_proof );
    weak_builder_r.WeakTo( rule->GetConclusion() );
    start_rule = weak_builder_r.BuildRule();

#ifdef DEBUG_PRINT_INFO
Print::PrintSequent( start_rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
#endif

    PermuationRuleBuilder order_perm_builder( start_rule, true, m_lk_proof );
    order_perm_builder.OrderLike( rule->GetConclusion() );
    start_rule = order_perm_builder.BuildRule();

    PermuationRuleBuilder order_perm_builder_r( start_rule, false, m_lk_proof );
    order_perm_builder_r.OrderLike( rule->GetConclusion() );
    start_rule = order_perm_builder_r.BuildRule();

#ifdef DEBUG_PRINT_INFO
Print::PrintSequent( start_rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
#endif
    if( remove_help_rule )
    {
      ASSERT( help_rule != NULL && help_rule->GetParent() != NULL );

#ifdef DEBUG_PRINT_INFO
      std::cerr << "removing help rule" << std::endl;
      std::cerr << "help_rule conclusion: ";
      Print::PrintSequent( help_rule->GetConclusion(), std::cerr );
      std::cerr << std::endl;

      std::cerr << "help_rule name: " << help_rule->GetName() << std::endl;
#endif

      ProofLib::Rule *r_t = help_rule->GetParent()->CastTo<ProofLib::Rule>();

#ifdef DEBUG_PRINT_INFO
      std::cerr << "help_rule parent (r_t) conclusion: ";
      Print::PrintSequent( r_t->GetConclusion(), std::cerr );
      std::cerr << std::endl;

      std::cerr << "help_rule parent (r_t) name: " << r_t->GetName() << std::endl;
#endif

      ASSERT( r_t != NULL );

      r_t->ReplaceRuleItem( help_rule, help_rule->GetRuleItem(0) );
      help_rule->RemoveAt(0);

      if( help_rule == start_rule )
      {
        start_rule = r_t;
      }
      
      delete help_rule;
    }
#if 0
    rule->ReplaceRuleItem( rule->GetRuleItem(0), start_rule );
#else
    if( rule != start_rule )
    {
#ifdef DEBUG_PRINT_INFO
      std::cerr << "removing another rule" << std::endl;
#endif
      rule->RemoveAt( 0 );
      delete rule;
      p->SetEndRule( start_rule );
    }
#endif

    // hack to fix the bug that caused an empty rule application
    // at the end of a proof
#if 0
    ProofLib::Sequent *parent = rule->GetRuleItem(0)->CastTo<ProofLib::Sequent>();
    ASSERT( parent != NULL );
    if ( rule->GetConclusion() == parent )
    {
      std::cerr << "conclusions are equal!" << std::endl;
    }
#endif
  }

  //////////////////////////////////////////////////////////////////////////////

  void AddSequentToWorld( ProofLib::Sequent* seq,
                          FormulaSet &world_l,
                          FormulaSet &world_r )
  {
    ASSERT( seq != NULL );

    ProofLib::SElementList *a = seq->GetLeft();
    bool on_left_side = true;

    // build the world
    while( true )
    {
      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();
#if 0
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
#endif

        // add to the world
        ( on_left_side ? world_l : world_r ).insert( f_a );
      }

      a = seq->GetRight();
      if( !on_left_side )
        break; // finally done

      on_left_side = false;
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  void ZipPredecessorsIntoSequent( ProofLib::Sequent *into,
                                   ProofLib::Sequent *from,
                                   int omitt_left,
                                   int omitt_right )
  {
    ASSERT( into != NULL );
    ASSERT( from != NULL );
    if( omitt_left < 0 ) omitt_left = 0;
    if( omitt_right < 0 ) omitt_right = 0;

    if( from->GetLeft() != NULL && from->GetLeft()->GetChildCount() > 0 )
    {
      if( into->GetLeft() == NULL )
        into->SetLeft( new ProofLib::SElementList() );

      ProofLib::SElementList *l = into->GetLeft();

      size_t i = from->GetLeft()->GetChildCount();

      while( i > size_t(omitt_left) )
        l->AddLeft( from->GetLeft()->GetSElement(--i)->Clone() );
    }

    if( from->GetRight() != NULL && from->GetRight()->GetChildCount() > 0 )
    {
      if( into->GetRight() == NULL )
        into->SetRight( new ProofLib::SElementList() );

      ProofLib::SElementList *l = into->GetRight();

      size_t i = from->GetRight()->GetChildCount() - omitt_right;

      while( i > 0 )
        l->AddLeft( from->GetRight()->GetSElement(--i)->Clone() );
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  static void SetRuleName( ProofLib::Rule *rule, int rule_name,
                           std::string s_param = std::string() )
  {
    USTRING name;
    name.FromUtf8( rule_names[rule_name].latex_code );
    LatexConvertor.KnownLatexSymbolsToUnicode( name );

    rule->SetName( name );

    USTRING type;
    type.FromUtf8( rule_names[rule_name].xml_name );

    rule->SetType( type );

    if( !s_param.empty() )
    {
      USTRING param;
      param.FromUtf8( s_param.c_str() );

      rule->SetParam( param );
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  class RuleStepInfo
  {
  public:
    RuleStepInfo()
    {
      rule_name = branch_no = left_new = right_new =
      left_principals = right_principals = -1;
    }

    int rule_name;
    int branch_no;
    int left_new;
    int right_new;
    int left_principals;
    int right_principals;
  };

  RuleStepInfo GetRuleStepInfo( const USTRING& type, size_t branch_no )
  {
    RuleStepInfo info;

    info.branch_no = branch_no;

    int rule_name = 0;

    std::string str;
    bool ret = UTF8FromUString( type, str );
    ASSERT( ret );

    while( rule_names[rule_name].xml_name != NULL )
    {
      if( !strcmp( str.c_str(), rule_names[rule_name].xml_name ) )
        break;
      rule_name++;
    }

    if( rule_names[rule_name].xml_name == NULL )
      return info;

    info.rule_name = rule_name;

    switch( rule_name )
    {
      case RULE_ANDR:
        ASSERT( branch_no == 0 || branch_no == 1 );
        info.left_new = 0;
        info.right_new = 1;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_ANDL1:
      case RULE_ANDL2:
        ASSERT( branch_no == 0 );
        info.left_new = 1;
        info.right_new = 0;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_ORL:
        ASSERT( branch_no == 0 || branch_no == 1 );
        info.left_new = 1;
        info.right_new = 0;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_ORR1:
      case RULE_ORR2:
        ASSERT( branch_no == 0 );
        info.left_new = 0;
        info.right_new = 1;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_IMPLL:
        ASSERT( branch_no == 0 || branch_no == 1 );
        info.left_new = branch_no == 0 ? 0 : 1;
        info.right_new = branch_no == 0 ? 1 : 0;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_IMPLR:
        ASSERT( branch_no == 0 );
        info.left_new = 1;
        info.right_new = 1;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_NEGR:
        ASSERT( branch_no == 0 );
        info.left_new = 1;
        info.right_new = 0;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_NEGL:
        ASSERT( branch_no == 0 );
        info.left_new = 0;
        info.right_new = 1;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_FORALLR:
      case RULE_EXISTSR:
      case RULE_FORALLR2:
      case RULE_EXISTSR2:
        ASSERT( branch_no == 0 );
        info.left_new = 0;
        info.right_new = 1;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_FORALLL:
      case RULE_EXISTSL:
      case RULE_FORALLL2:
      case RULE_EXISTSL2:
        ASSERT( branch_no == 0 );
        info.left_new = 1;
        info.right_new = 0;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_EQL1:
      case RULE_EQL2:
        ASSERT( branch_no == 0 || branch_no == 1 );
        info.left_new = branch_no == 0 ? 0 : 1;
        info.right_new = branch_no == 0 ? 1 : 0;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_EQR1:
      case RULE_EQR2:
        ASSERT( branch_no == 0 || branch_no == 1 );
        info.left_new = 0;
        info.right_new = 1;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_CUT:
        ASSERT( branch_no == 0 || branch_no == 1 );
        info.left_new = branch_no == 0 ? 0 : 1;
        info.right_new = branch_no == 0 ? 1 : 0;
        info.left_principals = 0;
        info.right_principals = 0;
        break;
      case RULE_UNDEFL:
        ASSERT( branch_no == 0 );
        info.left_new = 1;
        info.right_new = 0;
        info.left_principals = 1;
        info.right_principals = 0;
        break;
      case RULE_UNDEFR:
        ASSERT( branch_no == 0 );
        info.left_new = 0;
        info.right_new = 1;
        info.left_principals = 0;
        info.right_principals = 1;
        break;
      case RULE_WEAKR:
      case RULE_WEAKL:
      case RULE_CONTRR:
      case RULE_CONTRL:
      case RULE_PERMR:
      case RULE_PERML:
      default:
        ASSERT( false ); // what the hell happened here?
    }

    return info;
  }

  //////////////////////////////////////////////////////////////////////////////

  int FindFormulaIn( ProofLib::Formula *f, ProofLib::Sequent *seq, bool left,
                     size_t start = 0 )
  {
    ProofLib::SElementList *a = left ? seq->GetLeft() : seq->GetRight();

    if( a != NULL ) for( size_t i_a = start; i_a < a->GetChildCount(); i_a++ )
    {
      ProofLib::SFormula *sf_a =
        dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );

      ASSERT( sf_a != NULL );

      if( f == sf_a->GetFormula() )
        return i_a;
    }

    return -1;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* CompleteProof( ProofLib::Rule* rule,
                                 FormulaSet &world_l,
                                 FormulaSet &world_r,
                                 int parent_aux_l = -1, int parent_aux_r = -1 )
  {
    if( rule->GetRuleItemCount() == 0 )
    {
      return
        BuildAxiom( rule, world_l, world_r );
#if 0
      ProofLib::Sequent *seq =
        BuildAxiom( rule->GetConclusion(), world_l, world_r );

      if( seq == NULL )
        return rule;

      ProofLib::Sequent *t = rule->GetConclusion();
      rule->SetConclusion( seq );
      delete t;
      return rule;
#endif
    }

    ProofLib::Sequent *myseq = rule->GetConclusion();
    ProofLib::Sequent *newseq = new ProofLib::Sequent( NULL, NULL );

    bool unary_rule = true;

    ProofLib::Formula *f_new_l = NULL;
    ProofLib::Formula *f_new_r = NULL;
    ProofLib::Formula *f_pri_l = NULL;
    ProofLib::Formula *f_pri_r = NULL;

    // get information about this rule
    RuleStepInfo step_info = GetRuleStepInfo( rule->GetType(), 0 );

    if( myseq->GetLeft() != NULL && myseq->GetLeft()->GetChildCount() > 0 )
    {
      if( parent_aux_l > 0 )
      {
        ASSERT( parent_aux_l == 1 );

        // pointer to new left
        size_t i = myseq->GetLeft()->GetChildCount() - 1;
  
        ProofLib::SFormula *sf =
          dynamic_cast<ProofLib::SFormula*>( myseq->GetLeft()->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        f_new_l = sf->GetFormula();
        f_new_l->Attach();
      }

      // pointer to principal left
      if( step_info.left_principals > 0 )
      {
        ASSERT( step_info.left_principals == 1 );

        size_t i = 0;
  
        ProofLib::SFormula *sf =
          dynamic_cast<ProofLib::SFormula*>( myseq->GetLeft()->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        f_pri_l = sf->GetFormula();
        f_pri_l->Attach();
      }
    }

    if( myseq->GetRight() != NULL && myseq->GetRight()->GetChildCount() > 0 )
    {
      if( parent_aux_r > 0 )
      {
        ASSERT( parent_aux_r == 1 );
        size_t i = 0;
  
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( myseq->GetRight()->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        f_new_r = sf->GetFormula();
        f_new_r->Attach();
      }

      // pointer to principal right
      if( step_info.right_principals > 0 )
      {
        ASSERT( step_info.right_principals == 1 );

        size_t i = myseq->GetRight()->GetChildCount()-1;
  
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( myseq->GetRight()->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        f_pri_r = sf->GetFormula();
        f_pri_r->Attach();
      }
    }

    // add new formulas to the world.
    AddSequentToWorld( rule->GetConclusion(), world_l, world_r );

    delete myseq;

    // begin from right to left, since we use the world for the left branch
    for( int i = rule->GetRuleItemCount() - 1; i >= 0; i-- )
    {
      ProofLib::Sequent *seq = NULL;
      ProofLib::ProofLink *pl = NULL;
      ProofLib::Rule *r = NULL;
      ProofLib::Proof *this_p = NULL;

      // get information about this rule
      RuleStepInfo step_info = GetRuleStepInfo( rule->GetType(), i );

      if( (seq = rule->GetRuleItem(i)->CastTo<ProofLib::Sequent>()) != NULL )
      {
        // is this a 'explicit axiom' rule?
        if( rule->GetRuleItemCount() == 1 )
        {
          // comes from explicit axiom rule ...
          ZipPredecessorsIntoSequent( newseq, seq, 0, 0 );

          AddToAxioms( newseq );
          
          // just to revent destroying the sequent, when destroying the
          // rule later...
          rule->SetConclusion( new ProofLib::Sequent( NULL, NULL ) );
  
          ProofLib::Rule *new_rule = new ProofLib::Rule;
  
          ProofLib::Rule *parent = rule->GetParent()->CastTo<ProofLib::Rule>();
          if( parent != NULL )
            parent->ReplaceRuleItem( rule, new_rule );
          else
          {
            ASSERT( rule->GetParent() == NULL );

            m_lk_proof->SetEndRule( new_rule );
          }
  
          delete rule;
          rule = new_rule;
          break;
        }
        else
        {
          ProofLib::Sequent *seq_t = BuildAxiomSequent( seq, world_l, world_r );

          if( seq_t == NULL )
            AddToAxioms( seq );

          if( seq_t != NULL && seq_t != seq )
          {
            rule->ReplaceRuleItem( seq, seq_t );
            delete seq;
            seq = seq_t;
          }
        }
      }
      if( (pl = rule->GetRuleItem(i)->CastTo<ProofLib::ProofLink>()) != NULL )
      {
        r = pl->GetLinkedProof()->GetEndRule()->CastTo<ProofLib::Rule>();
        this_p = m_lk_proof;
        m_lk_proof = pl->GetLinkedProof();
      }
      if( r != NULL ||
          (r = rule->GetRuleItem(i)->CastTo<ProofLib::Rule>()) != NULL )
      {
        // rekursive call
        if( i == 0 ) //rule->GetRuleItemCount() == 1 )
          // use my world for the first branch
          r = CompleteProof( r, world_l, world_r,
                             step_info.left_new, step_info.right_new );
        else
        {
          // make copies of the world!
          FormulaSet new_world_l(world_l);
          FormulaSet new_world_r(world_r);

          r = CompleteProof( r, new_world_l, new_world_r,
                             step_info.left_new, step_info.right_new );
          unary_rule = false;
        }

        if( this_p != NULL )
          m_lk_proof = this_p;

        seq = r->GetConclusion();
      }
      
      ASSERT( seq != NULL );

      ZipPredecessorsIntoSequent( newseq, seq,
                                  step_info.left_new, step_info.right_new );
#if 0
      std::cerr << i << ": " << step_info.left_new << " - "
                << step_info.right_new;
      std::cerr << std::endl;
#endif
    }

#if 0
    Print::PrintSequent( newseq, std::cerr );
    std::cerr << std::endl;
#endif
    rule->SetConclusion( newseq );

    if( f_pri_l )
      newseq->AddLeft( f_pri_l );
    if( f_pri_r )
      newseq->AddRight( f_pri_r );

    rule = BuildStructuralRules( rule, f_new_l, f_pri_l, true, unary_rule );

    rule = BuildStructuralRules( rule, f_new_r, f_pri_r, false, unary_rule );

    if( f_pri_l != NULL ) f_pri_l->Detach();
    if( f_pri_r != NULL ) f_pri_r->Detach();
    if( f_new_l != NULL ) f_new_l->Detach();
    if( f_new_r != NULL ) f_new_r->Detach();

    return rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  /// builds permutation rule from top to bottom, i.e. 'from' is permise and 
  /// 'to' the conclusion of the perm rule.
  class PermuationRuleBuilder
  {
    ProofLib::Proof *p;
    ProofLib::Sequent *from, *to;
    ProofLib::Rule *r_from;
    bool left;
    size_t count;
    bool builded;

    typedef std::vector<size_t> Permuation;
  
    Permuation perm;

  public:
    PermuationRuleBuilder( ProofLib::Rule *from, bool left, ProofLib::Proof *p )
    {
      ASSERT( from != NULL );
      ASSERT( p != NULL );
      this->r_from = from;
      this->from = from->GetConclusion();
      this->to = this->from->Clone();
      this->left = left;
      this->p = p;

      if( left )
        count = to->GetLeft() == NULL ? 0 : to->GetLeft()->GetChildCount();
      else
        count = to->GetRight() == NULL ? 0 : to->GetRight()->GetChildCount();

      builded = false;
    }

    ProofLib::Proof* GetProof() { return p; }

    ProofLib::Rule* BuildRule()
    {
      ASSERT( !builded );

      builded = true;

      std::ostringstream str;

      std::vector<bool> done( perm.size(), false );

      bool no_perm = true;
      for( size_t i = 0; i < perm.size(); i++ )
        if( perm[i] != i && !done[i] )
        {
          size_t start = i;
          size_t j = i;
          done[start] = true;

          str << "(" << start + 1;

          while( perm[j] != j && perm[j] != start )
          {
            str << " " << perm[j] + 1;
            size_t k = j;
            j = perm[j];
            done[j] = true;
            perm[k] = k; // reset permutation
          }

          str << ")";
          no_perm = false;
        }

      if( no_perm )
      {
        delete to;
        return r_from;
      }

#if 0
std::cerr << str.str() << std::endl;
#endif
      ProofLib::Rule *perm_rule = new ProofLib::Rule;

      perm_rule->SetConclusion( to );
      SetRuleName( perm_rule, left ? RULE_PERML : RULE_PERMR, str.str() );

      ProofLib::Rule *parent = r_from->GetParent()->CastTo<ProofLib::Rule>();
      if( parent == NULL )
      {
        ASSERT( GetProof()->GetEndRule() == r_from );
        GetProof()->SetEndRule( perm_rule );
      }
      else
        parent->ReplaceRuleItem( r_from, perm_rule );

      perm_rule->AddLeft( r_from );

      return perm_rule;
    }

    void AddPermutation( size_t a, size_t b )
    {
#if 0 //def DEBUG_PRINT_INFO
std::cerr << "AddPermutation( " << a << ", " << b << " )" << std::endl;
#endif
      ASSERT( !builded );

      if( a == b )
        return;

#if 0
      if( a > b )
        std::swap( a, b );
  
      ASSERT( b < count );
        
      if( perm.size() <= b )
      {
        size_t i = perm.size();
        perm.resize( b + 1 );

        for( ; i <= b; i++ )
          perm[i] = i; // fill with identity
      }

      for( size_t i = 0; i < perm.size(); i++ )
        if( i != a )
        {
          ASSERT( i == b || perm[i] != b );
        }
        else
        {
          ASSERT( perm[i] == b || perm[i] == a );
        }
#endif

      ASSERT( a < count );
      ASSERT( b < count );

      size_t m = std::max( a, b );

      if( perm.size() <= m )
      {
        size_t i = perm.size();
        perm.resize( m + 1 );

        for( ; i <= m; i++ )
          perm[i] = i; // fill with identity
      }

      if( perm[a] == b )
        return;

      ASSERT( perm[a] == a );

      for( size_t i = 0; i < perm.size(); i++ )
        if( i != a )
        {
          ASSERT( i == b || perm[i] != b );
        }

      perm[a] = b;

      if( left )
        to->GetLeft()->ExchangeSElements( a, b );
      else
        to->GetRight()->ExchangeSElements( a, b );

      return;
    }

    void OrderEqualsTogether()
    {
      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();

      if( l == NULL )
        return;

      for( size_t i = 0; i < l->GetChildCount() - 1; i++ )
      {
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        ProofLib::Formula *f = sf->GetFormula();

        for( size_t j = i + 1; j < l->GetChildCount(); j++ )
        {
          ProofLib::SFormula *sf_j =
           dynamic_cast<ProofLib::SFormula*>( l->GetSElement(j) );
    
          ASSERT( sf_j != NULL );
    
          if( f == sf_j->GetFormula() )
          {
            if( j == i + 1 )
              i++;
            else
              AddPermutation( ++i, j );
          }
        }
      }
    }

    void OrderEqualsTogetherRightToLeft()
    {
      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();

      if( l == NULL )
        return;

      for( int i = l->GetChildCount() - 1; i > 0; i-- )
      {
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        ProofLib::Formula *f = sf->GetFormula();

        for( int j = i - 1; j >= 0; j-- )
        {
          ProofLib::SFormula *sf_j =
           dynamic_cast<ProofLib::SFormula*>( l->GetSElement(j) );
    
          ASSERT( sf_j != NULL );
    
          if( f == sf_j->GetFormula() )
          {
            if( j == i - 1 )
              i--;
            else
              AddPermutation( --i, j );
          }
        }
      }
    }

    void OrderLike( ProofLib::Sequent *seq )
    {
#ifdef DEBUG_PRINT_INFO
std::cerr << "FROM: ";
Print::PrintSequent( to, std::cerr );
std::cerr << std::endl;
std::cerr << "TO: ";
Print::PrintSequent( seq, std::cerr );
std::cerr << std::endl;
#endif
      if( count == 0 )
        return;

      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();
      ProofLib::SElementList *l_g = left ? seq->GetLeft() : seq->GetRight();

      ASSERT( l->GetChildCount() == l_g->GetChildCount() );

      std::vector<bool> occupied( l->GetChildCount(), false );
      perm.resize( l->GetChildCount() );

      for( size_t i = 0; i < l->GetChildCount(); i++ )
        perm[i] = i;

      for( size_t i = 0; i < l->GetChildCount(); i++ )
      {
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        ProofLib::Formula *f = sf->GetFormula();

        bool found = false;
        for( size_t j = 0; j < l_g->GetChildCount(); j++ )
        {
          if( occupied[j] )
            continue;

          ProofLib::SFormula *sf_j =
           dynamic_cast<ProofLib::SFormula*>( l_g->GetSElement(j) );
    
          ASSERT( sf_j != NULL );
    
#if 0 //def DEBUG_PRINT_INFO
std::cerr << "HAVE at " << i << ": ";
Print::PrintFormula( f, std::cerr );
std::cerr << std::endl;
std::cerr << "FOUND at " << j << ": ";
Print::PrintFormula( sf_j->GetFormula(), std::cerr );
std::cerr << std::endl;
#endif
          if( f == sf_j->GetFormula() )
          {
            perm[j] = i;
            occupied[j] = true;
            found = true;
            break;
          }
        }

        ASSERT( found );
      }

#if 0
for( size_t i = 0; i < l->GetChildCount(); i++ )
  std::cerr << perm[i] << " ";
std::cerr << std::endl;
#endif

      if( left )
        to->SetLeft( l_g->Clone() );
      else
        to->SetRight( l_g->Clone() );
    }
  };

  //////////////////////////////////////////////////////////////////////////////

  /// builds contraction rule from top to bottom, i.e. 'from' is permise and 
  /// 'to' the conclusion of the contraction rule.
  class ContractionRuleBuilder
  {
    ProofLib::Proof *p;
    ProofLib::Sequent *from, *to;
    ProofLib::Rule *r_from;
    bool left;
    size_t count;
    bool builded;

    typedef std::vector<size_t> Contractions;
  
    Contractions cont;

  public:
    ContractionRuleBuilder( ProofLib::Rule *from, bool left,
                            ProofLib::Proof *p )
    {
#if 0
        std::cerr << "NEW ContractionRuleBuilder" << std::endl;
        std::cerr << "for ";
        std::cerr << ( left ? "left" : "right" );
        std::cerr << " of ";
        Print::PrintSequent( from->GetConclusion(), std::cerr );
        std::cerr << std::endl;
#endif
      ASSERT( from != NULL );
      ASSERT( p != NULL );
      this->r_from = from;
      this->from = from->GetConclusion();
      this->to = this->from->Clone();
      this->left = left;
      this->p = p;

      if( left )
        count = to->GetLeft() == NULL ? 0 : to->GetLeft()->GetChildCount();
      else
        count = to->GetRight() == NULL ? 0 : to->GetRight()->GetChildCount();

      builded = false;
    }

    ProofLib::Proof* GetProof() { return p; }

    ProofLib::Rule* BuildRule()
    {
      ASSERT( !builded );

      if( cont.size() == 0 )
        return r_from;

      ApplyContractions();

      builded = true;

      std::ostringstream str;

      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();

      bool no_cont = true;
      for( size_t i = 0; i < l->GetChildCount(); i++ )
        if( i < cont.size() )
        {
          if( i != 0 )
            str << ",";

          str << cont[i]; 

          if( cont[i] > 1 )
            no_cont = false;
        }
        else
        {
          ASSERT( i != 0 );
          str << ",1";
        }

      if( no_cont )
      {
        delete to;
        return r_from;
      }
#if 0
        std::cerr << str.str();
        std::cerr << std::endl;
#endif

      ProofLib::Rule *cont_rule = new ProofLib::Rule;

      cont_rule->SetConclusion( to );
      SetRuleName( cont_rule, left ? RULE_CONTRL : RULE_CONTRR, str.str() );

      ProofLib::Rule *parent = r_from->GetParent()->CastTo<ProofLib::Rule>();
      if( parent == NULL )
      {
        ASSERT( GetProof()->GetEndRule() == r_from );
        GetProof()->SetEndRule( cont_rule );
      }
      else
        parent->ReplaceRuleItem( r_from, cont_rule );

      cont_rule->AddLeft( r_from );

      return cont_rule;
    }

    void AddContraction( size_t n, size_t times )
    {
#if 0
        std::cerr << "AddContraction( " << n << ", " << times << " )";
        std::cerr << std::endl;
#endif
      ASSERT( !builded );
      ASSERT( times > 0 );
      ASSERT( n < count );

      if( cont.size() <= n )
      {
        size_t i = cont.size();
        cont.resize( n+1 );
        for( ; i <= n; i++ )
          cont[i] = 1; // fill with one
      }

      count -= cont[n];
      cont[n] = times;
      count += cont[n];
  
      ASSERT( n < count );
    }

    void ContractEquals()
    {
#if 0
std::cerr << "Contract Equals in " << std::endl;
Print::PrintSequent( to, std::cerr );
std::cerr << std::endl;
#endif

      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();

      if( l == NULL )
        return;

      size_t n = 0;
      for( size_t i = 0; i < l->GetChildCount() - 1; i++ )
      {
#if 0
        std::cerr << "i: " << i;
        std::cerr << std::endl;
#endif
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        ProofLib::Formula *f = sf->GetFormula();
        size_t times = 1;

        for( size_t j = i + 1; j < l->GetChildCount(); j++ )
        {
          ProofLib::SFormula *sf_j =
           dynamic_cast<ProofLib::SFormula*>( l->GetSElement(j) );
    
          ASSERT( sf_j != NULL );
    
          if( f == sf_j->GetFormula() )
          {
            times++;
            i++;
          }
          else
          {
            i = j - 1;
            break;
          }
        }
#if 0
        std::cerr << "i after: " << i << " times: " << times;
        std::cerr << std::endl;
#endif

        AddContraction( n, times );
        n++;
      }
    }

    void ContractTo( ProofLib::Sequent *like )
    {
#if 0
        std::cerr << "Contract" << std::endl;
        Print::PrintSequent( to, std::cerr );
        std::cerr << " to " << std::endl;
        std::cerr << ( left ? "left" : "right" );
        std::cerr << " of ";
        Print::PrintSequent( like, std::cerr );
        std::cerr << std::endl;
#endif
      ASSERT( to != NULL );

      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();

      if( l == NULL )
        return;

      size_t n = 0;
      for( size_t i = 0; i < l->GetChildCount() - 1; i++ )
      {
        ProofLib::SFormula *sf =
         dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i) );
  
        ASSERT( sf != NULL );
  
        ProofLib::Formula *f = sf->GetFormula();
        size_t times_0 = 1;

        for( size_t j = i + 1; j < l->GetChildCount(); j++ )
        {
          ProofLib::SFormula *sf_j =
           dynamic_cast<ProofLib::SFormula*>( l->GetSElement(j) );
    
          ASSERT( sf_j != NULL );
    
          if( f == sf_j->GetFormula() )
            times_0++;
          else
          {
            i = j-1;
            break;
          }
        }

        size_t times = like->CountOfFormula( f, left );

#if 0
        Print::PrintFormula( f, std::cerr );
        std::cerr << ": " << times_0;
        std::cerr << " -> " << times;
        std::cerr << std::endl;
#endif
    
        ASSERT( times_0 >= times );

        if( times > 0 )
          AddContraction( n, times_0 - times + 1 );
        n += times;
      }
    }

  private:
    void ApplyContractions()
    {
#if 0
std::cerr << "ApplyContractions" << std::endl;
#endif
      size_t c = count;

      if( cont.size() < count )
        c = cont.size();

      ProofLib::SElementList *l = left ? to->GetLeft() : to->GetRight();

      for( size_t i = 0; i < c; i++ )
      {
#if 0
std::cerr << "i: " << i << std::endl;
std::cerr << "cont[i]: " << cont[i] << std::endl;
std::cerr << "j_0: " << i+1;
std::cerr << std::endl;
std::cerr << "j_max:" << i + cont[i] - 1;
std::cerr << std::endl;
#endif
#ifdef DEBUG
        ProofLib::SFormula *sf_i =
          dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i) );
    
        ASSERT( sf_i != NULL );
#endif

        for( size_t j = i + 1; j < i + cont[i]; j++ )
        {
#if 0
std::cerr << "j:" << j;
std::cerr << std::endl;
#endif
#ifdef DEBUG
          ProofLib::SFormula *sf_j =
            dynamic_cast<ProofLib::SFormula*>( l->GetSElement(i+1) );

          ASSERT( sf_j != NULL );
    
          ASSERT( sf_i->GetFormula() == sf_j->GetFormula() );
#endif
          l->RemoveAt( i + 1 );
        }
      }
    }
  };

  //////////////////////////////////////////////////////////////////////////////

  /// builds weaking rule from top to bottom, i.e. 'from' is permise and 
  /// 'to' the conclusion of the weaking rule.
  class WeakeningRuleBuilder
  {
    ProofLib::Proof *p;
    ProofLib::Sequent *from, *to;
    ProofLib::Rule *r_from;
    bool left;
    size_t count;
    bool builded;

    typedef std::vector<ProofLib::Formula*> Weakens;
  
    Weakens weakens;

  public:
    WeakeningRuleBuilder( ProofLib::Rule *from, bool left, ProofLib::Proof *p )
    {
      ASSERT( from != NULL );
      this->r_from = from;
      this->from = from->GetConclusion();
      this->to = this->from->Clone();
      this->left = left;
      this->p = p;

      if( left )
        count = to->GetLeft() == NULL ? 0 : to->GetLeft()->GetChildCount();
      else
        count = to->GetRight() == NULL ? 0 : to->GetRight()->GetChildCount();

      builded = false;
    }

    ProofLib::Proof* GetProof() { return p; }

    ProofLib::Rule* BuildRule()
    {
      ASSERT( !builded );

      if( weakens.size() == 0 )
        return r_from;

      ProofLib::Rule *weak_rule = new ProofLib::Rule;

      for( size_t i = 0; i < weakens.size(); i++ )
      {
        if( left )
          to->AddLeft( weakens[i] );
        else
          to->AddRight( weakens[i] );
      }

      weak_rule->SetConclusion( to );

      SetRuleName( weak_rule, left ? RULE_WEAKL : RULE_WEAKR );

      ProofLib::Rule *parent = r_from->GetParent()->CastTo<ProofLib::Rule>();
      if( parent == NULL )
      {
        ASSERT( GetProof()->GetEndRule() == r_from );
        GetProof()->SetEndRule( weak_rule );
      }
      else
        parent->ReplaceRuleItem( r_from, weak_rule );

      weak_rule->AddLeft( r_from );

      return weak_rule;
    }

    void AddFormula( ProofLib::Formula *f )
    {
      weakens.push_back( f );
    }

    void WeakTo( ProofLib::Sequent *seq )
    {
      ProofLib::SElementList *l_to = left ? to->GetLeft() : to->GetRight();
      ProofLib::SElementList *l_from = left ? seq->GetLeft() : seq->GetRight();

      if( l_from == NULL )
        return;

      if( l_to == NULL )
      {
        for( size_t i = 0; i < l_from->GetChildCount(); i++ )
        {
          ProofLib::SFormula *sf =
            l_from->GetSElement(i)->CastTo<ProofLib::SFormula>();
          ASSERT( sf != NULL );

          ProofLib::Formula *f = sf->GetFormula();
          ASSERT( f != NULL );

          AddFormula( f );
        }

        return;
      }

      std::vector<ProofLib::SElement*> map( l_to->GetChildCount() + 1, NULL );

      for( size_t i = 0; i < l_from->GetChildCount(); i++ )
      {
        ProofLib::SFormula *sf =
          l_from->GetSElement(i)->CastTo<ProofLib::SFormula>();
        ASSERT( sf != NULL );

        ProofLib::Formula *f = sf->GetFormula();
        ASSERT( f != NULL );

        size_t j = 0;
        bool found = false;

        do
        {
          if( found ) j++;
          if( l_to == NULL || !l_to->FindFormula( f, j ) )
            break;
          else
            found = true;
        }
        while( map[j] != NULL );

        if( !found || j >= l_to->GetChildCount() || map[j] != NULL )
          AddFormula( f );
        else
          map[j] = sf;
      }
    }
  };

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* BuildStructuralRules(
    ProofLib::Rule *rule,
    ProofLib::Formula *f_parent_aux,
    ProofLib::Formula *f_principal,
    bool left,
    bool unary_rule )
  {
    ProofLib::Sequent *seq = rule->GetConclusion();

#ifdef DEBUG_PRINT_INFO
std::cerr << "BuildStructuralRules:" << std::endl;
Print::PrintSequent( seq, std::cerr );
std::cerr << std::endl;
std::cerr << "rule is " << (unary_rule ? "" : "not ") << "unary." << std::endl;
std::cerr << "application on " << ( left ? "left" : "right" ) << " side.";
std::cerr << std::endl;
if( f_parent_aux != NULL )
{
  std::cerr << "auxilary formula: ";
  Print::PrintFormula( f_parent_aux, std::cerr );
  std::cerr << std::endl;
}
if( f_principal != NULL )
{
  std::cerr << "principal formula: ";
  Print::PrintFormula( f_principal, std::cerr );
  std::cerr << std::endl;
}
#endif

//    if( !unary_rule ) // why only for binary rules???
    {
      PermuationRuleBuilder perm_builder( rule, left, m_lk_proof );
      if( left )
        perm_builder.OrderEqualsTogether();
      else
        perm_builder.OrderEqualsTogetherRightToLeft();
      rule = perm_builder.BuildRule();

#ifdef DEBUG_PRINT_INFO
Print::PrintSequent( rule->GetConclusion(), std::cerr );
std::cerr << std::endl;
#endif
      ContractionRuleBuilder cont_builder( rule, left, m_lk_proof );
      cont_builder.ContractEquals();
      rule = cont_builder.BuildRule();
      seq = rule->GetConclusion();
    }

#ifdef DEBUG_PRINT_INFO
std::cerr << "After Contraction/Permutation:" << std::endl;
Print::PrintSequent( seq, std::cerr );
std::cerr << std::endl;
#endif

    size_t count;

    if( left )
      count = seq->GetLeft() == NULL ? 0 : seq->GetLeft()->GetChildCount();
    else
      count = seq->GetRight() == NULL ? 0 : seq->GetRight()->GetChildCount();
    
    if( unary_rule && f_principal != NULL )
    {
      int pos = FindFormulaIn( f_principal, seq, left, left ? 1 : 0 );

      if( !left && size_t(pos) == count - 1 )
        pos = -1;

      // need permutation?
      if( ( left && pos > 1 ) ||
          ( !left &&  pos >= 0 &&  size_t(pos) < count - 2 ) )
      {
        PermuationRuleBuilder perm_builder( rule, left, m_lk_proof );
        perm_builder.AddPermutation( pos, left ? 1 : count - 2 );
        rule = perm_builder.BuildRule();
        seq = rule->GetConclusion();
      }

      // need contraction?
/*
      if( pos >= 0 )
      {
        ContractionRuleBuilder cont_builder( rule, left, m_lk_proof );
        cont_builder.AddContraction( left ? 0 : count - 2, 2 );
        rule = cont_builder.BuildRule();
        seq = rule->GetConclusion();
        count--;
      }
*/
    }

    if( f_parent_aux != NULL )
    {
      int pos = FindFormulaIn( f_parent_aux, seq, left );

#ifdef DEBUG_PRINT_INFO
Print::PrintSequent( seq, std::cerr );
std::cerr << std::endl;
#endif

      // need permuation?
      if( ( left && pos > 0 ) ||
          ( !left &&  pos >= 0 &&  size_t(pos) < count - 1 ) )
      {
        PermuationRuleBuilder perm_builder( rule, left, m_lk_proof );
        perm_builder.AddPermutation( pos, left ? 0 : count - 1 );
        rule = perm_builder.BuildRule();
        seq = rule->GetConclusion();
      }
      
      if( pos < 0 )
      {
        // need weaking
        WeakeningRuleBuilder weak_builder( rule, left, m_lk_proof );
        weak_builder.AddFormula( f_parent_aux );
        rule = weak_builder.BuildRule();
      }
    }

    return rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* BuildAxiom( ProofLib::Rule* rule,
                              FormulaSet &world_l,
                              FormulaSet &world_r )
  {
    ASSERT( rule != NULL );

    ProofLib::Sequent *seq =
      BuildAxiomSequent( rule->GetConclusion(), world_l, world_r );

    if( seq == NULL )
    {
      AddToAxioms( rule->GetConclusion() );
      return rule;
    }

    ProofLib::Sequent *t = rule->GetConclusion();
    rule->SetConclusion( seq );

    WeakeningRuleBuilder weak_builder_l( rule, true, m_lk_proof );
    weak_builder_l.WeakTo( t );
    rule = weak_builder_l.BuildRule();

    WeakeningRuleBuilder weak_builder_r( rule, false, m_lk_proof );
    weak_builder_r.WeakTo( t );
    rule = weak_builder_r.BuildRule();

    delete t;
    return rule;
  }

  ProofLib::Sequent* BuildAxiomSequent( ProofLib::Sequent* seq,
                                        FormulaSet &world_l,
                                        FormulaSet &world_r )
  {
    ASSERT( seq != NULL );
    
    ProofLib::Formula *f = NULL;

    ProofLib::SElementList *a = seq->GetLeft();
    bool on_left_side = true;

    // try to create an regular LK axiom
    while( true )
    {
      FormulaSet &world_d = ( !on_left_side ? world_l : world_r );

      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();
#if 0
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
#endif

        // is the dual formula in the world?
        if( world_d.find( f_a ) != world_d.end() )
        {
          f = f_a;
          break;
        }
      }

      a = seq->GetRight();
      if( f != NULL || !on_left_side )
        break; // finally done

      on_left_side = false;
    }
    
    // ok, if we didn't find an axiom by using the sequent formulas,
    // we try now to find on in the world an use this.
    if( f == NULL ) for( FormulaSet::const_iterator i = world_l.begin();
                         i != world_l.end();
                         i++ )
    {
      if( world_r.find( *i ) != world_r.end() )
      {
        f = *i;
        break;
      }
    }

    // at least we check if there is an equation which serves as an axiom.
    if( f == NULL && seq->GetRight() != NULL )
    {
      for( size_t i = 0; i < seq->GetRight()->GetChildCount(); i++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( seq->GetRight()->GetSElement(i) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::AtomFormula *a =
          sf_a->GetFormula()->CastTo<ProofLib::AtomFormula>();

        if( a == NULL )
          continue;

        const PredicateSymbol *ps =
          dynamic_cast<const PredicateSymbol*>( a->GetSymbol() );

        ASSERT( ps != NULL );

        if( ps->IsEqualPred() )
        {
          seq = new ProofLib::Sequent( NULL, NULL );
          seq->AddRight( a );
          AddToAxioms( seq );
          return seq;
        }
      }
    }

    if( f == NULL )
    {
#ifdef DEBUG_PRINT_INFO
std::cerr << "initial sequent which is no regular axiom: ";
Print::PrintSequent( seq, std::cerr );
std::cerr << std::endl;
#endif
      return NULL;
    }

    seq = new ProofLib::Sequent( NULL, NULL );
    seq->AddLeft( f );
    seq->AddRight( f );

    AddToAxioms( seq );
    return seq;
  }

  //////////////////////////////////////////////////////////////////////////////

  void AddToAxioms( ProofLib::Sequent* seq )
  {
    std::ostream &out = *Transformer::axiom_writer;
    bool regular_axiom = false;
    ProofLib::Sequent* schema = NULL;

    // check for A :- A type
    if( seq->GetLeft() != NULL &&
        seq->GetRight() != NULL &&
        seq->GetLeft()->GetChildCount() == 1 &&
        seq->GetRight()->GetChildCount() == 1 )
    {
      ProofLib::SFormula *sf_a =
        dynamic_cast<ProofLib::SFormula*>( seq->GetLeft()->GetSElement(0) );
      ProofLib::SFormula *sf_b =
        dynamic_cast<ProofLib::SFormula*>( seq->GetRight()->GetSElement(0) );

      ASSERT( sf_a != NULL );
      ASSERT( sf_b != NULL );

      regular_axiom = ( sf_a->GetFormula() == sf_b->GetFormula() );

      if( sf_a->GetFormula()->CastTo<ProofLib::AtomFormula>() == NULL )
      {
        std::cerr << "WARNING: Have a non atomar regular axiom in proof "
                  << m_lk_proof->GetSymbol()->GetName()
                  << ": ";
        Print::PrintSequent( seq, std::cerr );
        std::cerr << std::endl;
      }
    }
    else
    // check for :- s = s type 
    if( ( seq->GetLeft() == NULL || seq->GetLeft()->GetChildCount() == 0 )
          &&
        ( seq->GetRight() != NULL && seq->GetRight()->GetChildCount() == 1 ) )
    {
      ProofLib::SFormula *sf =
        dynamic_cast<ProofLib::SFormula*>( seq->GetRight()->GetSElement(0) );
      ASSERT( sf != NULL );

      ProofLib::AtomFormula *af =
        sf->GetFormula()->CastTo<ProofLib::AtomFormula>();

      if( af != NULL )
      {
        const PredicateSymbol *ps =
          dynamic_cast<const PredicateSymbol*>( af->GetSymbol() );

        regular_axiom = ( ps->IsEqualPred() &&
                          af->GetTerm(0) == af->GetTerm(1) );
      }
    }

    // check for defined axiom
    if( !regular_axiom )
    {
      AxiomTable *at = GetAxiomTable();

      for( size_t i = 0; i < at->GetCount(); i++ )
      {
        ProofLib::CheckInstanceOperation op;

        if( op.IsInstanceOf( seq, at->GetAxiomAt(i) ) )
        {
          schema = at->GetAxiomAt(i);
          break;
        }
      }

      if( schema == NULL )
      {
        std::cerr << "WARNING: Have a non regular axiom in proof "
                  << m_lk_proof->GetSymbol()->GetName()
                  << " which does not fit to any declared axiom schema: ";
        Print::PrintSequent( seq, std::cerr );
        std::cerr << std::endl;
      }
    }

    if( Transformer::axiom_writer != NULL )
    {
      if( regular_axiom )
        out << "REGULAR AXIOM";
      else
        out << "AXIOM";
  
      out << " (in proof "
          << m_lk_proof->GetSymbol()->GetName()
          << "): ";
  
      Print::PrintSequent( seq, out );

      if( schema != NULL )
      {
        out << " instance of ";
        Print::PrintSequent( schema, out );
      }

      out << std::endl;
    }
  }
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

std::ostream *Transformer::axiom_writer = NULL;

Transformer* Transformer::NewTransformer( AxiomTable* axiomtbl )
{
  return new TransformerImpl( axiomtbl );
}

////////////////////////////////////////////////////////////////////////////////

}
