////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: typehlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include "typehlk.h"
#include "symbolhlk.h"
#include "proofhlk.h"
#include "printhlk.h"

#include "libproof/libproof.h"
#include "util/hash.h"
#include "util/myassert.h"

namespace HLK
{

class AxiomTableImpl :
  public AxiomTable,
  private std::vector< ProofLib::Sequent* >
//private stlhash::hash_set< ProofLib::Sequent* >
{
private:
  typedef stlhash::hash_map< USTRING, ProofLib::Sequent*>
          StringToSequent;

  StringToSequent named;

public:
  virtual ~AxiomTableImpl() {}

  virtual void AddAxiom( ProofLib::Sequent *seq, const wchar_t *name = NULL )
  {
    size_t i = 0;
    for( ; i < size(); i++ )
      if( (*this)[i] == seq )
        break;

    if( i == size() )
      push_back( seq );

    if( name != NULL )
    {
      if( named.find( name ) != named.end() && named[name] != seq )
      {
        GenericException *e = new GenericException;
        e->Stream() << "Axiom '" << name << "' already exists, defined as ";
        Print::PrintSequent( seq, e->Stream() );
        e->Stream() << ".";
        throw e;
      }
      else
        named[name] = seq;
    }
  }

  virtual size_t GetCount() const
  {
    return size();
  }

  virtual ProofLib::Sequent* GetAxiomAt( size_t index ) const
  {
    ASSERT( index < size() );
    return *(begin() + index);
  }

  virtual const char* GetNameAt( size_t index ) const
  {
    ASSERT( false );
    return NULL;
  }

  virtual ProofLib::Sequent* GetAxiom( const wchar_t *name )
  {
    StringToSequent::iterator i = named.find( name );

    if( i == named.end() )
      return NULL;

    return i->second;
  }
};

////////////////////////////////////////////////////////////////////////////////

Document::Document()
{
  m_proof_table = new ProofTable;
  m_axiom_table = new AxiomTableImpl;
  m_symbol_table = new SymbolTable;
  m_node_database = new ProofLib::NodeDatabase;
}

Document::~Document()
{
  delete m_proof_table;
  delete m_axiom_table;
  delete m_node_database;
  delete m_symbol_table;
}

SymbolTable* Document::GetSymbolTable() const
{
  return m_symbol_table;
}

ProofTable* Document::GetProofTable() const
{
  return m_proof_table;
}

AxiomTable* Document::GetAxiomTable() const
{
  return m_axiom_table;
}

}
